from base64 import urlsafe_b64encode
from hashlib import sha1
from typing import List, TYPE_CHECKING


if TYPE_CHECKING:
    from snowddl.engine import SnowDDLEngine


class SnowDDLQueryBuilder:
    def __init__(self, engine: "SnowDDLEngine"):
        self.engine = engine
        self.fragments: List[List[str]] = [[]]

    def append(self, sql, params=None):
        sql = self.engine.format(sql, params)
        self.fragments[-1].append(sql)

    def append_nl(self, sql, params=None):
        sql = self.engine.format(sql, params)
        self.fragments.append([sql])

    def fragment_count(self):
        return sum([len(f) for f in self.fragments])

    def add_short_hash(self, comment):
        if comment:
            return f"{comment} {self._short_hash()}"

        return self._short_hash()

    def compare_short_hash(self, comment):
        if comment is None:
            return False

        return str(comment).endswith(self._short_hash())

    def _short_hash(self):
        sha1_digest = sha1(str(self).encode('UTF-8')).digest()
        return f"#{urlsafe_b64encode(sha1_digest[:12]).decode('ascii')}"

    def __str__(self):
        return '\n'.join([' '.join(line) for line in self.fragments])



from snowddl.blueprint import ProcedureBlueprint, Ident, IdentWithPrefix, ComplexIdentWithPrefixAndArgs, NameWithType, DataType
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


procedure_json_schema = {
    "type": "object",
    "properties": {
        "language": {
            "type": "string"
        },
        "body": {
            "type": "string"
        },
        "arguments": {
            "type": "object",
            "additionalProperties": {
                "type": "string"
            }
        },
        "returns": {
            "type": "string"
        },
        "is_strict": {
            "type": "boolean"
        },
        "is_execute_as_caller": {
            "type": "boolean"
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["body", "returns"],
    "additionalProperties": False
}


class ProcedureParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("procedure", procedure_json_schema, self.process_procedure)

    def process_procedure(self, f: ParsedFile):
        arguments = [NameWithType(name=Ident(k), type=DataType(t)) for k, t in f.params.get('arguments', {}).items()]
        base_name = self.validate_name_with_args(f.path, arguments)

        bp = ProcedureBlueprint(
            full_name=ComplexIdentWithPrefixAndArgs(self.env_prefix, f.database, f.schema, base_name, data_types=[a.type.base_type for a in arguments]),
            database=IdentWithPrefix(self.env_prefix, f.database),
            schema=Ident(f.schema),
            name=Ident(base_name),
            language=f.params.get('language', 'SQL'),
            body=f.params['body'],
            arguments=arguments,
            returns=DataType(f.params['returns']),
            is_strict=f.params.get('is_strict', False),
            is_immutable=f.params.get('is_immutable', False),
            is_execute_as_caller=f.params.get('is_execute_as_caller', False),
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import ExternalFunctionBlueprint
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import dtypes_from_arguments


class ExternalFunctionResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.EXTERNAL_FUNCTION

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW EXTERNAL FUNCTIONS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            full_name = f"{r['catalog_name']}.{r['schema_name']}.{r['name']}({dtypes_from_arguments(r['arguments'])})"

            existing_objects[full_name] = {
                "database": r['catalog_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "arguments": r['arguments'],
                "comment": r['description'],
                "is_secure": r['is_secure'] == 'Y',
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ExternalFunctionBlueprint)

    def create_object(self, bp: ExternalFunctionBlueprint):
        query = self._build_create_function(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl("COMMENT ON FUNCTION {full_name:i} IS {comment}", {
            "full_name": bp.full_name,
            "comment": query.add_short_hash(bp.comment),
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: ExternalFunctionBlueprint, row: dict):
        query = self._build_create_function(bp)

        if not query.compare_short_hash(row['comment']):
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl("COMMENT ON FUNCTION {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            })

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl("DROP FUNCTION {database:i}.{schema:i}.{name:i}({dtypes:r})", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
            "dtypes": dtypes_from_arguments(row['arguments']),
        })

        return ResolveResult.DROP

    def _build_create_function(self, bp: ExternalFunctionBlueprint):
        query = self.engine.query_builder()
        query.append("CREATE OR REPLACE")

        if bp.is_secure:
            query.append("SECURE")

        query.append("EXTERNAL FUNCTION {database:i}.{schema:i}.{name:i} (", {
            "database": bp.database,
            "schema": bp.schema,
            "name": bp.name,
        })

        for idx, arg in enumerate(bp.arguments):
            query.append_nl("    {comma:r}{arg_name:i} {arg_type:r}", {
                "comma": "  " if idx == 0 else ", ",
                "arg_name": arg.name,
                "arg_type": arg.type,
            })

        query.append_nl(")")

        query.append_nl("RETURNS {ret_type:r}", {
            "ret_type": bp.returns,
        })

        query.append_nl("API_INTEGRATION = {api_integration:i}", {
            "api_integration": bp.api_integraton,
        })

        if bp.is_strict:
            query.append_nl("STRICT")

        if bp.is_immutable:
            query.append_nl("IMMUTABLE")

        if bp.headers:
            query.append_nl("HEADERS = (")

            for idx, (header_name, header_value) in enumerate(bp.headers.items()):
                query.append_nl("    {comma:r}{header_name} = {header_value}", {
                    "comma": "  " if idx == 0 else ", ",
                    "header_name": header_name,
                    "header_value": header_value,
                })

            query.append_nl(")")

        if bp.context_headers:
            query.append_nl("CONTEXT_HEADERS = ({context_headers:r})", {
                "context_headers": bp.context_headers,
            })

        if bp.max_batch_rows:
            query.append_nl("MAX_BATCH_ROWS = {max_batch_rows:d}", {
                "max_batch_rows": bp.max_batch_rows,
            })

        if bp.compression:
            query.append_nl("COMPRESSION = {compression}", {
                "compression": bp.compression,
            })

        if bp.request_translator:
            query.append_nl("REQUEST_TRANSLATOR = {request_translator:i}", {
                "request_translator": bp.request_translator,
            })

        if bp.response_translator:
            query.append_nl("RESPONSE_TRANSLATOR = {response_translator:i}", {
                "response_translator": bp.response_translator,
            })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        query.append_nl("AS {url}", {
            "url": bp.url,
        })

        return query



from snowddl.blueprint import DatabaseBlueprint, IdentWithPrefix
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


database_json_schema = {
    "type": "object",
    "properties": {
        "is_transient": {
            "type": "boolean"
        },
        "retention_time": {
            "type": "integer"
        },
        "is_sandbox": {
            "type": "boolean"
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}


class DatabaseParser(AbstractParser):
    def load_blueprints(self):
        for database_path in self.base_path.iterdir():
            if not database_path.is_dir():
                continue

            params = self.parse_single_file(database_path / 'params.yaml', database_json_schema)

            bp = DatabaseBlueprint(
                full_name=IdentWithPrefix(self.env_prefix, database_path.name),
                database = IdentWithPrefix(self.env_prefix, database_path.name),
                is_transient=params.get('is_transient', False),
                retention_time=params.get('retention_time', None),
                is_sandbox=params.get('is_sandbox', False),
                comment=params.get('comment', None),
            )

            self.config.add_blueprint(bp)



from snowddl.blueprint import FileFormatBlueprint, Ident, IdentWithPrefix, ComplexIdentWithPrefix
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


file_format_json_schema = {
    "type": "object",
    "properties": {
        "type": {
            "type": "string"
        },
        "format_options": {
            "type": "object",
            "additionalProperties": {
                "type": ["array", "boolean", "number", "string"]
            }
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["type"],
    "additionalProperties": False
}


class FileFormatParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("file_format", file_format_json_schema, self.process_file_format)

    def process_file_format(self, f: ParsedFile):
        bp = FileFormatBlueprint(
            full_name=ComplexIdentWithPrefix(self.env_prefix, f.database, f.schema, f.name),
            database=IdentWithPrefix(self.env_prefix, f.database),
            schema=Ident(f.schema),
            name=Ident(f.name),
            type=f.params['type'].upper(),
            format_options={option_name.upper(): option_value for option_name, option_value in f.params.get('format_options', {}).items()},
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)



from snowddl.blueprint import ProcedureBlueprint, ComplexIdentWithPrefixAndArgs
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import dtypes_from_arguments


class ProcedureResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.PROCEDURE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW USER PROCEDURES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            full_name = f"{r['catalog_name']}.{r['schema_name']}.{r['name']}({dtypes_from_arguments(r['arguments'])})"

            existing_objects[full_name] = {
                "database": r['catalog_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "arguments": r['arguments'],
                "comment": r['description'],
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ProcedureBlueprint)

    def create_object(self, bp: ProcedureBlueprint):
        query = self._build_create_procedure(bp)

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl("COMMENT ON PROCEDURE {full_name:i} IS {comment}", {
            "full_name": bp.full_name,
            "comment": query.add_short_hash(bp.comment),
        })

        return ResolveResult.CREATE

    def compare_object(self, bp: ProcedureBlueprint, row: dict):
        query = self._build_create_procedure(bp)

        if not query.compare_short_hash(row['comment']):
            self.engine.execute_safe_ddl(query)

            self.engine.execute_safe_ddl("COMMENT ON PROCEDURE {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": query.add_short_hash(bp.comment),
            })

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl("DROP PROCEDURE {database:i}.{schema:i}.{name:i}({dtypes:r})", {
            "database": row['database'],
            "schema": row['schema'],
            "name": row['name'],
            "dtypes": dtypes_from_arguments(row['arguments']),
        })

        return ResolveResult.DROP

    def _build_create_procedure(self, bp: ProcedureBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE OR REPLACE PROCEDURE {database:i}.{schema:i}.{name:i} (", {
            "database": bp.database,
            "schema": bp.schema,
            "name": bp.name,
        })

        for idx, arg in enumerate(bp.arguments):
            query.append_nl("    {comma:r}{arg_name:i} {arg_type:r}", {
                "comma": "  " if idx == 0 else ", ",
                "arg_name": arg.name,
                "arg_type": arg.type,
            })

        query.append_nl(")")

        query.append_nl("RETURNS {ret_type:r}", {
            "ret_type": bp.returns,
        })

        query.append_nl("LANGUAGE {language:r}", {
            "language": bp.language,
        })

        if bp.is_strict:
            query.append_nl("STRICT")

        if bp.is_immutable:
            query.append_nl("IMMUTABLE")

        if bp.is_execute_as_caller:
            query.append_nl("EXECUTE AS CALLER")

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        query.append_nl("AS {body}", {
            "body": bp.body,
        })

        return query



from snowddl.blueprint import TagBlueprint, Ident, IdentWithPrefix, ComplexIdentWithPrefix, ObjectType, TagReference
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


tag_json_schema = {
    "type": "object",
    "properties": {
        "references": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "object_type": {
                        "type": "string"
                    },
                    "object_name": {
                        "type": "string"
                    },
                    "column_name": {
                        "type": "string"
                    },
                    "tag_value": {
                        "type": "string"
                    }
                },
                "required": ["object_type", "object_name", "tag_value"],
                "additionalProperties": False
            },
            "minItems": 1
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}


class TagParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("tag", tag_json_schema, self.process_tag)

    def process_tag(self, f: ParsedFile):
        references = []

        for a in f.params.get('references', []):
            ref = TagReference(
                object_type=ObjectType[a['object_type'].upper()],
                object_name=self.config.build_complex_ident(a['object_name'], f.database, f.schema),
                column_name=Ident(a['column_name']) if a.get('column_name') else None,
                tag_value=a['tag_value'],
            )

            references.append(ref)

        bp = TagBlueprint(
            full_name=ComplexIdentWithPrefix(self.env_prefix, f.database, f.schema, f.name),
            database=IdentWithPrefix(self.env_prefix, f.database),
            schema=Ident(f.schema),
            name=Ident(f.name),
            references=references,
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from json import loads
from jsonschema import validate
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

from snowddl.config import SnowDDLConfig
from snowddl.blueprint import ComplexIdentWithPrefix, ComplexIdentWithPrefixAndArgs, NameWithType
from snowddl.parser._parsed_file import ParsedFile


class AbstractParser(ABC):
    def __init__(self, config: SnowDDLConfig, base_path: Path):
        self.config = config
        self.base_path = base_path

        self.env_prefix = config.env_prefix

    @abstractmethod
    def load_blueprints(self):
        pass

    def parse_single_file(self, path: Path, json_schema: dict, callback: Callable[[ParsedFile],Union[None,Dict]] = None):
        if not callback:
            callback = lambda f: f.params

        if path.exists():
            try:
                file = ParsedFile(self, path, json_schema)
                return callback(file)
            except Exception as e:
                self.config.add_error(path, e)

        return {}

    def parse_schema_object_files(self, object_type: str, json_schema: dict, callback: Callable[[ParsedFile],None]):
        for path in self.base_path.glob(f"*/*/{object_type}/*.yaml"):
            try:
                file = ParsedFile(self, path, json_schema)
                callback(file)
            except Exception as e:
                self.config.add_error(path, e)

    def normalise_params_dict(self, params):
        if params is None:
            return None

        if isinstance(params, dict):
            return {k.upper(): v for k, v in params.items()}

        raise ValueError(f"Value is neither None, nor dictionary [{params}]")

    def validate_name_with_args(self, path: Path, arguments: List[NameWithType]):
        stem_name = str(path.stem)
        args_str = ','.join([a.type.base_type.name for a in arguments]).lower()

        open_pos = stem_name.find('(')
        close_pos = stem_name.find(')')

        if open_pos == -1 or close_pos == -1:
            raise ValueError(f"File [{path}] name should have list of arguments, e.g. [{stem_name}({args_str}).yaml]")

        base_name = stem_name[:open_pos]

        if stem_name[open_pos+1:close_pos] != args_str:
            raise ValueError(f"File [{path}] name does not match list of arguments, expected [{base_name}({args_str}).yaml]")

        return base_name



from .abc_resolver import AbstractResolver, ResolveResult
from .abc_role_resolver import AbstractRoleResolver
from .abc_schema_object_resolver import AbstractSchemaObjectResolver
from .account_params import AccountParameterResolver
from .business_role import BusinessRoleResolver
from .database import DatabaseResolver
from .external_function import ExternalFunctionResolver
from .external_table import ExternalTableResolver
from .file_format import FileFormatResolver
from .foreign_key import ForeignKeyResolver
from .function import FunctionResolver
from .masking_policy import MaskingPolicyResolver
from .materialized_view import MaterializedViewResolver
from .network_policy import NetworkPolicyResolver
from .pipe import PipeResolver
from .primary_key import PrimaryKeyResolver
from .procedure import ProcedureResolver
from .resource_monitor import ResourceMonitorResolver
from .row_access_policy import RowAccessPolicyResolver
from .sequence import SequenceResolver
from .schema import SchemaResolver
from .schema_role import SchemaRoleResolver
from .stage import StageResolver
from .stage_file import StageFileResolver
from .stream import StreamResolver
from .table import TableResolver
from .task import TaskResolver
from .tech_role import TechRoleResolver
from .view import ViewResolver
from .unique_key import UniqueKeyResolver
from .user import UserResolver
from .user_role import UserRoleResolver
from .warehouse import WarehouseResolver
from .warehouse_role import WarehouseRoleResolver


default_resolver_sequence = [
    AccountParameterResolver,
    NetworkPolicyResolver,
    ResourceMonitorResolver,
    WarehouseResolver,
    WarehouseRoleResolver,
    DatabaseResolver,
    SchemaResolver,
    SchemaRoleResolver,
    FileFormatResolver,
    StageResolver,
    StageFileResolver,
    SequenceResolver,
    FunctionResolver,
    ExternalFunctionResolver,
    ProcedureResolver,
    TableResolver,
    ExternalTableResolver,
    PrimaryKeyResolver,
    UniqueKeyResolver,
    ForeignKeyResolver,
    StreamResolver,
    MaterializedViewResolver,
    ViewResolver,
    PipeResolver,
    TaskResolver,
    MaskingPolicyResolver,
    RowAccessPolicyResolver,
    TechRoleResolver,
    BusinessRoleResolver,
    UserRoleResolver,
    UserResolver,
]



from enum import Enum


class ObjectType(Enum):
    ACCOUNT_PARAMETER = {
        "singular": "ACCOUNT PARAMETER",
        "plural": "ACCOUNT PARAMETERS",
    }

    DATABASE = {
        "singular": "DATABASE",
        "plural": "DATABASES",
    }

    EXTERNAL_FUNCTION = {
        "singular": "EXTERNAL FUNCTION",
        "plural": "EXTERNAL FUNCTIONS",
        "is_future_grant_supported": True,
    }

    EXTERNAL_TABLE = {
        "singular": "EXTERNAL TABLE",
        "plural": "EXTERNAL TABLES",
        "simplified": "TABLE",
        "is_future_grant_supported": True,
    }

    FILE_FORMAT = {
        "singular": "FILE FORMAT",
        "plural": "FILE FORMATS",
        "is_future_grant_supported": True,
    }

    FUNCTION = {
        "singular": "FUNCTION",
        "plural": "FUNCTIONS",
        "is_future_grant_supported": True,
    }

    MASKING_POLICY = {
        "singular": "MASKING POLICY",
        "plural": "MASKING POLICIES",
    }

    MATERIALIZED_VIEW = {
        "singular": "MATERIALIZED VIEW",
        "plural": "MATERIALIZED VIEWS",
        "simplified": "VIEW",
        "is_future_grant_supported": True,
    }

    NETWORK_POLICY = {
        "singular": "NETWORK POLICY",
        "plural": "NETWORK POLICIES",
    }

    PIPE = {
        "singular": "PIPE",
        "plural": "PIPES",
        "is_future_grant_supported": True,
    }

    PROCEDURE = {
        "singular": "PROCEDURE",
        "plural": "PROCEDURES",
        "is_future_grant_supported": True,
    }

    RESOURCE_MONITOR = {
        "singular": "RESOURCE MONITOR",
        "plural": "RESOURCE MONITORS",
    }

    ROLE = {
        "singular": "ROLE",
        "plural": "ROLES",
    }

    ROW_ACCESS_POLICY = {
        "singular": "ROW ACCESS POLICY",
        "plural": "ROW ACCESS POLICIES",
    }

    SEQUENCE = {
        "singular": "SEQUENCE",
        "plural": "SEQUENCES",
        "is_future_grant_supported": True,
    }

    SCHEMA = {
        "singular": "SCHEMA",
        "plural": "SCHEMAS",
        "is_future_grant_supported": True,
    }

    STAGE = {
        "singular": "STAGE",
        "plural": "STAGES",
        "is_future_grant_supported": True,
    }

    STAGE_FILE = {
        "singular": "STAGE FILE",
        "plural": "STAGE FILES",
    }

    STREAM = {
        "singular": "STREAM",
        "plural": "STREAMS",
        "is_future_grant_supported": True,
    }

    TABLE = {
        "singular": "TABLE",
        "plural": "TABLES",
        "is_future_grant_supported": True,
    }

    TAG = {
        "singular": "TAG",
        "plural": "TAGS",
        "is_future_grant_supported": True,
    }

    TASK = {
        "singular": "TASK",
        "plural": "TASKS",
        "is_future_grant_supported": True,
    }

    VIEW = {
        "singular": "VIEW",
        "plural": "VIEWS",
        "is_future_grant_supported": True,
    }

    USER = {
        "singular": "USER",
        "plural": "USERS",
    }

    WAREHOUSE = {
        "singular": "WAREHOUSE",
        "plural": "WAREHOUSES",
    }

    # Constraints
    PRIMARY_KEY = {
        "singular": "PRIMARY KEY",
        "plural": "PRIMARY KEYS",
    }

    UNIQUE_KEY = {
        "singular": "UNIQUE KEY",
        "plural": "UNIQUE KEYS",
    }

    FOREIGN_KEY = {
        "singular": "FOREIGN KEY",
        "plural": "FOREIGN KEYS",
    }

    @property
    def singular(self):
        return self.value.get('singular')

    @property
    def plural(self):
        return self.value.get('plural')

    @property
    def simplified(self):
        return self.value.get('simplified', self.value.get('singular'))

    @property
    def is_future_grant_supported(self) -> bool:
        return self.value.get('is_future_grant_supported', False)

    def __repr__(self):
        return f"<{self.__class__.__name__}.{super().name}>"



from snowddl.blueprint import ViewBlueprint

from snowddl.error import SnowDDLExecuteError
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType


class ViewResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.VIEW

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW VIEWS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            if r['is_materialized'] == 'true':
                continue

            existing_objects[f"{r['database_name']}.{r['schema_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "text": r['text'],
                "is_secure": r['is_secure'] == "true",
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(ViewBlueprint)

    def create_object(self, bp: ViewBlueprint):
        self.engine.execute_safe_ddl(self._build_create_view(bp))

        # Comments on views are broken and must be applied separately
        if bp.comment:
            self.engine.execute_safe_ddl("COMMENT ON VIEW {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": bp.comment,
            })

        return ResolveResult.CREATE

    def compare_object(self, bp: ViewBlueprint, row: dict):
        query = self._build_create_view(bp)

        # If view text is exactly the same
        if row['text'] == str(query):
            try:
                # ... and it is possible to query view (underlying objects were not changed)
                self.engine.describe_meta("SELECT * FROM {full_name:i}", {
                    "full_name": bp.full_name,
                })
            except SnowDDLExecuteError as e:
                # TODO: add checks for specific errors?
                pass
            else:
                # Comments on views are broken and must be applied separately
                if bp.comment != row['comment']:
                    self.engine.execute_safe_ddl("COMMENT ON VIEW {full_name:i} IS {comment}", {
                        "full_name": bp.full_name,
                        "comment": bp.comment if bp.comment else '',
                    })

                    return ResolveResult.ALTER
                else:
                    return ResolveResult.NOCHANGE

        # Replace view if we got here
        self.engine.execute_safe_ddl(query)

        # Comments on views are broken and must be applied separately
        if bp.comment:
            self.engine.execute_safe_ddl("COMMENT ON VIEW {full_name:i} IS {comment}", {
                "full_name": bp.full_name,
                "comment": bp.comment,
            })

        return ResolveResult.REPLACE

    def drop_object(self, row: dict):
        self.engine.execute_safe_ddl("DROP VIEW {database:i}.{schema:i}.{view_name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "view_name": row['name'],
        })

        return ResolveResult.DROP

    def _build_create_view(self, bp: ViewBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE OR REPLACE")

        if bp.is_secure:
            query.append("SECURE")

        query.append("VIEW {database:i}.{schema:i}.{view_name:i}", {
            "database": bp.database,
            "schema": bp.schema,
            "view_name": bp.name,
        })

        if bp.columns:
            query.append_nl('(')

            for idx, c in enumerate(bp.columns):
                query.append_nl("    {comma:r}{col_name:i}", {
                    "comma": "  " if idx == 0 else ", ",
                    "col_name": c.name,
                })

                if c.comment:
                    query.append("COMMENT {col_comment}", {
                        "col_comment": c.comment,
                    })

            query.append_nl(')')

        query.append_nl("COPY GRANTS")
        query.append_nl("AS")
        query.append_nl(bp.text)

        return query



from collections import defaultdict
from pathlib import Path
from traceback import TracebackException
from typing import List, Dict, Type, Union, TYPE_CHECKING

from snowddl.blueprint import AbstractBlueprint, T_Blueprint, BaseDataType, Ident, IdentWithPrefix, ComplexIdentWithPrefix, ComplexIdentWithPrefixAndArgs


class SnowDDLConfig:
    BUSINESS_ROLE_SUFFIX = 'B_ROLE'
    SCHEMA_ROLE_SUFFIX = 'S_ROLE'
    TECH_ROLE_SUFFIX = 'T_ROLE'
    USER_ROLE_SUFFIX = 'U_ROLE'
    WAREHOUSE_ROLE_SUFFIX = 'W_ROLE'

    def __init__(self, env_prefix=None):
        self.env_prefix = self._init_env_prefix(env_prefix)

        self.blueprints: Dict[Type[T_Blueprint], Dict[str,T_Blueprint]] = defaultdict(dict)
        self.errors: List[dict] = []

        self.placeholders: Dict[str,Union[bool,float,int,str]] = {}

    def get_blueprints_by_type(self, cls: Type[T_Blueprint]) -> Dict[str,T_Blueprint]:
        return self.blueprints.get(cls, {})

    def get_placeholder(self, name: str) -> Union[bool,float,int,str]:
        if name not in self.placeholders:
            raise ValueError(f"Unknown placeholder [{name}]")

        return self.placeholders[name]

    def add_blueprint(self, bp: AbstractBlueprint):
        self.blueprints[bp.__class__][str(bp.full_name)] = bp

    def add_error(self, path: Path, e: Exception):
        self.errors.append({
            "path": path,
            "error": e,
        })

    def add_placeholder(self, name: str, value: Union[bool,float,int,str]):
        self.placeholders[name] = value

    def build_complex_ident(self, object_name, context_database_name=None, context_schema_name=None) -> ComplexIdentWithPrefix:
        # Function or procedure identifier with arguments
        if object_name.endswith(')'):
            object_name, data_types_str = object_name.rstrip(')').split('(')
            data_types = [BaseDataType[dt.strip(' ').upper()] for dt in data_types_str.split(',')] if data_types_str else []
        else:
            data_types = None

        parts = object_name.split('.')
        parts_len = len(parts)

        if parts_len > 3:
            raise ValueError(f"Too many delimiters in schema object identifier [{object_name}]")

        # Add context db_name to "schema_name.object_name" or "object_name" identifier
        if parts_len <= 2 and context_database_name:
            parts.insert(0, context_database_name)

        # Add context schema_name to "object_name" identifier
        if parts_len == 1 and context_schema_name:
            parts.insert(1, context_schema_name)

        # Function or procedure identifier with arguments
        if isinstance(data_types, list):
            return ComplexIdentWithPrefixAndArgs(self.env_prefix, *parts, data_types=data_types)
        else:
            return ComplexIdentWithPrefix(self.env_prefix, *parts)

    def build_role_ident(self, *args: Union[Ident, str]) -> IdentWithPrefix:
        parts = [str(a.value) if isinstance(a, Ident) else str(a) for a in args]
        return IdentWithPrefix(self.env_prefix, f"{'__'.join(parts)}")

    def _init_env_prefix(self, env_prefix):
        if env_prefix:
            env_prefix = str(env_prefix).upper()

            if '__' in env_prefix:
                raise ValueError(f"Env prefix [{env_prefix}] cannot contain [__] double underscore")

            if env_prefix.endswith('_'):
                raise ValueError(f"Env prefix [{env_prefix}] cannot end with [_] underscore")

            return f"{env_prefix}__"

        return ''



from snowddl.blueprint import SchemaRoleBlueprint, SchemaBlueprint, Grant, FutureGrant
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class SchemaRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.SCHEMA_ROLE_SUFFIX

    def get_blueprints(self):
        blueprints = []

        for schema in self.config.get_blueprints_by_type(SchemaBlueprint).values():
            blueprints.append(self.get_blueprint_owner_role(schema))
            blueprints.append(self.get_blueprint_read_role(schema))
            blueprints.append(self.get_blueprint_write_role(schema))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_owner_role(self, schema: SchemaBlueprint):
        grants = []
        future_grants = []

        grants.append(Grant(
            privilege="USAGE",
            on=ObjectType.DATABASE,
            name=schema.database,
        ))

        grants.append(Grant(
            privilege="USAGE",
            on=ObjectType.SCHEMA,
            name=schema.full_name,
        ))

        create_object_types = [
            ObjectType.FUNCTION,
            ObjectType.PROCEDURE,
            ObjectType.TABLE,
            ObjectType.VIEW,
        ]

        for object_type in create_object_types:
            grants.append(Grant(
                privilege=f"CREATE {object_type.singular}",
                on=ObjectType.SCHEMA,
                name=schema.full_name,
            ))

        ownership_object_types = [
            ObjectType.EXTERNAL_TABLE,
            ObjectType.FILE_FORMAT,
            ObjectType.FUNCTION,
            ObjectType.MATERIALIZED_VIEW,
            ObjectType.PIPE,
            ObjectType.PROCEDURE,
            ObjectType.SEQUENCE,
            ObjectType.STREAM,
            ObjectType.TABLE,
            ObjectType.TASK,
            ObjectType.VIEW,
        ]

        for object_type in ownership_object_types:
            future_grants.append(FutureGrant(
                privilege="OWNERSHIP",
                on=object_type,
                name=schema.full_name,
            ))

        privileges_map = {
            ObjectType.STAGE: ['READ', 'WRITE', 'USAGE'],
        }

        for object_type, privileges in privileges_map.items():
            for privilege in privileges:
                future_grants.append(FutureGrant(
                    privilege=privilege,
                    on=object_type,
                    name=schema.full_name,
                ))

        depends_on = []

        for additional_grant in schema.owner_additional_grants:
            # Dependency on another schema role
            if additional_grant.on == ObjectType.ROLE and str(additional_grant.name).endswith(self.get_role_suffix()):
                depends_on.append(additional_grant.name)

            grants.append(additional_grant)

        bp = SchemaRoleBlueprint(
            full_name=self.config.build_role_ident(schema.database, schema.schema, 'OWNER', self.get_role_suffix()),
            grants=grants,
            future_grants=future_grants,
            comment=None,
            depends_on=depends_on,
        )

        return bp

    def get_blueprint_read_role(self, schema: SchemaBlueprint):
        grants = []
        future_grants = []

        grants.append(Grant(
            privilege="USAGE",
            on=ObjectType.DATABASE,
            name=schema.database,
        ))

        grants.append(Grant(
            privilege="USAGE",
            on=ObjectType.SCHEMA,
            name=schema.full_name,
        ))

        privileges_map = {
            ObjectType.EXTERNAL_TABLE: ['SELECT', 'REFERENCES'],
            ObjectType.FILE_FORMAT: ['USAGE'],
            ObjectType.FUNCTION: ['USAGE'],
            ObjectType.MATERIALIZED_VIEW: ['SELECT', 'REFERENCES'],
            ObjectType.PROCEDURE: ['USAGE'],
            ObjectType.STAGE: ['READ', 'USAGE'],
            ObjectType.STREAM: ['SELECT'],
            ObjectType.TABLE: ['SELECT', 'REFERENCES'],
            ObjectType.VIEW: ['SELECT', 'REFERENCES'],
        }

        for object_type, privileges in privileges_map.items():
            for privilege in privileges:
                future_grants.append(FutureGrant(
                    privilege=privilege,
                    on=object_type,
                    name=schema.full_name,
                ))

        bp = SchemaRoleBlueprint(
            full_name=self.config.build_role_ident(schema.database, schema.schema, 'READ', self.get_role_suffix()),
            grants=grants,
            future_grants=future_grants,
            comment=None,
            depends_on=[],
        )

        return bp

    def get_blueprint_write_role(self, schema: SchemaBlueprint):
        grants = []
        future_grants = []

        grants.append(Grant(
            privilege="USAGE",
            on=ObjectType.DATABASE,
            name=schema.database,
        ))

        grants.append(Grant(
            privilege="USAGE",
            on=ObjectType.SCHEMA,
            name=schema.full_name,
        ))

        privileges_map = {
            ObjectType.STAGE: ['READ', 'WRITE'],
            ObjectType.SEQUENCE: ['USAGE'],
            ObjectType.TABLE: ['INSERT', 'UPDATE', 'DELETE', 'TRUNCATE'],
        }

        for object_type, privileges in privileges_map.items():
            for privilege in privileges:
                future_grants.append(FutureGrant(
                    privilege=privilege,
                    on=object_type,
                    name=schema.full_name,
                ))

        bp = SchemaRoleBlueprint(
            full_name=self.config.build_role_ident(schema.database, schema.schema, 'WRITE', self.get_role_suffix()),
            grants=grants,
            future_grants=future_grants,
            comment=None,
            depends_on=[],
        )

        return bp



from logging import getLogger, NullHandler
from pathlib import Path
from snowflake.connector import SnowflakeConnection
from traceback import TracebackException
from typing import Optional

from snowddl.config import SnowDDLConfig
from snowddl.converter import default_converter_sequence
from snowddl.engine import SnowDDLEngine
from snowddl.parser import default_parser_sequence, PlaceholderParser
from snowddl.resolver import default_resolver_sequence
from snowddl.settings import SnowDDLSettings
from snowddl.version import __version__


logger = getLogger(__name__)
logger.addHandler(NullHandler())


class SnowDDLApp:
    def __init__(self):
        self.config: Optional[SnowDDLConfig] = None
        self.engine: Optional[SnowDDLEngine] = None

        self.logger = logger

    def init_config(self, env_prefix):
        self.config = SnowDDLConfig(env_prefix)

    def init_engine(self, connection: SnowflakeConnection, settings: SnowDDLSettings):
        self.engine = SnowDDLEngine(connection, self.config, settings)

    def load_placeholders_with_parsers(self, base_path: Path, placeholder_path: Path = None):
        if placeholder_path and not placeholder_path.is_file():
            raise ValueError(f"Invalid placeholders path [{placeholder_path}]")

        parser = PlaceholderParser(self.config, base_path, placeholder_path)
        parser.load_blueprints()

    def load_blueprints_with_parsers(self, base_path: Path):
        if not base_path.is_dir():
            raise ValueError(f"Invalid config directory [{base_path}]")

        for parser_cls in default_parser_sequence:
            parser = parser_cls(self.config, base_path)
            parser.load_blueprints()

    def resolve_objects(self):
        with self.engine as engine:
            for resolver_cls in default_resolver_sequence:
                resolver = resolver_cls(engine)
                resolver.resolve()

    def destroy_objects(self):
        with self.engine as engine:
            for resolver_cls in default_resolver_sequence:
                resolver = resolver_cls(engine)
                resolver.destroy()

            self.engine.context.destroy_role_with_prefix()

    def convert_objects(self, base_path: Path):
        with self.engine as engine:
            for converter_cls in default_converter_sequence:
                converter = converter_cls(engine, base_path)
                converter.convert()

    def output_context(self):
        roles = []

        if self.engine.context.is_account_admin:
            roles.append("ACCOUNTADMIN")

        if self.engine.context.is_sys_admin:
            roles.append("SYSADMIN")

        if self.engine.context.is_security_admin:
            roles.append("SECURITYADMIN")

        self.logger.info(f"Snowflake version = {self.engine.context.version} ({self.engine.context.edition.name}), SnowDDL version = {__version__}")
        self.logger.info(f"Session = {self.engine.context.current_session}, User = {self.engine.context.current_user}")
        self.logger.info(f"Role = {self.engine.context.current_role}, Warehouse = {self.engine.context.current_warehouse}")
        self.logger.info(f"Roles in session = {','.join(roles)}")
        self.logger.info("---")

    def output_config_errors(self):
        for e in self.config.errors:
            self.logger.warning(f"[{e['path']}]: {''.join(TracebackException.from_exception(e['error']).format())}")

    def output_engine_stats(self):
        self.logger.info(f"Executed {len(self.engine.executed_ddl)} DDL queries, Suggested {len(self.engine.suggested_ddl)} DDL queries")

    def output_suggested_ddl(self):
        if self.engine.suggested_ddl:
            print("--- Suggested DDL ---\n")

        for sql in self.engine.suggested_ddl:
            print(f"{sql};\n")

    def output_executed_ddl(self):
        if self.engine.executed_ddl:
            print("--- Executed DDL ---\n")

        for sql in self.engine.executed_ddl:
            print(f"{sql};\n")



from snowddl.blueprint import TagBlueprint, ObjectType
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult


class TagResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True

    def get_object_type(self) -> ObjectType:
        return ObjectType.TAG

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW TAGS IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"

            existing_objects[full_name] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(TagBlueprint)

    def create_object(self, bp: TagBlueprint):
        self._create_tag(bp)
        self._apply_tag_refs(bp)

        return ResolveResult.CREATE

    def compare_object(self, bp: TagBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if self._apply_tag_refs(bp):
            result = ResolveResult.ALTER

        if row['comment'] != bp.comment:
            self.engine.execute_unsafe_ddl("ALTER TAG {full_name:i} SET COMMENT = {comment}", {
                "full_name": bp.full_name,
                "comment": bp.comment,
            })

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self._drop_tag_refs(row['database'], row['schema'], row['name'])
        self._drop_tag(row['database'], row['schema'], row['name'])

        return ResolveResult.DROP

    def _create_tag(self, bp: TagBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE TAG {full_name:i}", {
            "full_name": bp.full_name,
        })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_unsafe_ddl(query)

    def _drop_tag(self, database, schema, name):
        self.engine.execute_unsafe_ddl("DROP TAG {database:i}.{schema:i}.{name:i}", {
            "database": database,
            "schema": schema,
            "name": name,
        })

    def _apply_tag_refs(self, bp: TagBlueprint):
        existing_tag_refs = self._get_existing_tag_refs(bp.database, bp.schema, bp.name)
        applied_change = False

        for ref in bp.references:
            if ref.column_name:
                ref_key = f"{ref.object_type.name}|{ref.object_name}|{ref.column_name}|{bp.full_name}"
            else:
                ref_key = f"{ref.object_type.name}|{ref.object_name}|{bp.full_name}"

            # Tag was applied before
            if ref_key in existing_tag_refs:
                existing_tag_value = existing_tag_refs[ref_key]['tag_value']
                del(existing_tag_refs[ref_key])

                # Tag was applied with the same value
                if ref.tag_value == existing_tag_value:
                    continue

            # Apply tag
            if ref.column_name:
                self.engine.execute_unsafe_ddl("ALTER {object_type:r} {object_name:i} MODIFY COLUMN {column_name:i} SET TAG {tag_name:i} = {tag_value}", {
                    "object_type": ref.object_type.singular,
                    "object_name": ref.object_name,
                    "column_name": ref.column_name,
                    "tag_name": bp.full_name,
                })
            else:
                self.engine.execute_unsafe_ddl("ALTER {object_type:r} {object_name:i} SET TAG {tag_name:i} = {tag_value}", {
                    "object_type": ref.object_type.singular,
                    "object_name": ref.object_name,
                    "tag_name": bp.full_name,
                })

            applied_change = True

        # Remove remaining tag references which no longer exist in blueprint
        for existing_ref in existing_tag_refs.values():
            if existing_ref['column_name']:
                self.engine.execute_unsafe_ddl("ALTER {object_type:r} {object_name:i} MODIFY COLUMN {column_name:i} UNSET TAG {tag_name:i}", {
                    "object_type": existing_ref['object_type'],
                    "database": existing_ref['database'],
                    "schema": existing_ref['schema'],
                    "name": existing_ref['name'],
                    "column_name": existing_ref['column_name'],
                    "tag_name": bp.full_name,
                })
            else:
                self.engine.execute_unsafe_ddl("ALTER {object_type:r} {object_name:i} UNSET TAG {tag_name:i}", {
                    "object_type": existing_ref['object_type'],
                    "database": existing_ref['database'],
                    "schema": existing_ref['schema'],
                    "tag_name": bp.full_name,
                })

            applied_change = True

        return applied_change

    def _drop_tag_refs(self, database, schema, name):
        existing_policy_refs = self._get_existing_policy_refs(database, schema, name)

        for existing_ref in existing_policy_refs.values():
            if existing_ref['column_name']:
                self.engine.execute_unsafe_ddl("ALTER {object_type:r} {database:i}.{schema:i}.{name:i} MODIFY COLUMN {column_name:i} UNSET TAG {tag_database:i}.{tag_schema:i}.{tag_name:i}", {
                    "object_type": existing_ref['object_type'],
                    "database": existing_ref['database'],
                    "schema": existing_ref['schema'],
                    "name": existing_ref['name'],
                    "column_name": existing_ref['column_name'],
                    "tag_database": database,
                    "tag_schema": schema,
                    "tag_name": name,
                })
            else:
                self.engine.execute_unsafe_ddl("ALTER {object_type:r} {database:i}.{schema:i}.{name:i} UNSET TAG {tag_database:i}.{tag_schema:i}.{tag_name:i}", {
                    "object_type": existing_ref['object_type'],
                    "database": existing_ref['database'],
                    "schema": existing_ref['schema'],
                    "tag_database": database,
                    "tag_schema": schema,
                    "tag_name": name,
                })

    def _get_existing_tag_refs(self, database, schema, name):
        existing_policy_refs = {}

        # TODO: discover a better way to get tag references in real time
        # Currently it is not clear how to get all tag references properly

        return existing_policy_refs



from snowddl.blueprint import WarehouseBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType
from snowddl.resolver._utils import coalesce


class WarehouseResolver(AbstractResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.WAREHOUSE

    def get_existing_objects(self):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW WAREHOUSES LIKE {env_prefix:ls}", {
            'env_prefix': self.config.env_prefix,
        })

        for r in cur:
            if r['owner'] != self.engine.context.current_role:
                continue

            existing_objects[r['name']] = {
                "name": r['name'],
                "size": r['size'],
                "min_cluster_count": r.get('min_cluster_count', None),
                "max_cluster_count": r.get('max_cluster_count', None),
                "scaling_policy": r.get('scaling_policy', None),
                "auto_suspend": r['auto_suspend'],
                "resource_monitor": r['resource_monitor'] if r['resource_monitor'] != 'null' else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(WarehouseBlueprint)

    def create_object(self, bp: WarehouseBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE WAREHOUSE {name:i}", {
            "name": bp.full_name,
        })

        query.append_nl(self._build_common_properties(bp))
        query.append_nl("AUTO_RESUME = TRUE")
        query.append_nl("INITIALLY_SUSPENDED = TRUE")

        self.engine.execute_safe_ddl(query)

        if bp.resource_monitor:
            self.engine.execute_safe_ddl("ALTER WAREHOUSE {full_name:i} SET RESOURCE_MONITOR = {resource_monitor:i}", {
                "full_name": bp.full_name,
                "resource_monitor": bp.resource_monitor,
            }, condition=self.engine.settings.execute_resource_monitor)

        return ResolveResult.CREATE

    def compare_object(self, bp: WarehouseBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if self._compare_properties(bp, row):
            result = ResolveResult.ALTER

        if self._compare_resource_monitor(bp, row):
            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP WAREHOUSE {name:i}", {
            "name": row['name'],
        })

        return ResolveResult.DROP

    def _compare_properties(self, bp: WarehouseBlueprint, row: dict):
        if self._normalise_warehouse_size(bp.size) == self._normalise_warehouse_size(row['size']) \
            and bp.auto_suspend == row['auto_suspend'] \
            and (row['min_cluster_count'] is None or coalesce(bp.min_cluster_count, 1) == row['min_cluster_count']) \
            and (row['max_cluster_count'] is None or coalesce(bp.max_cluster_count, 1) == row['max_cluster_count']) \
            and (row['scaling_policy'] is None or coalesce(bp.scaling_policy, 'STANDARD') == row['scaling_policy']):
            return False

        query = self.engine.query_builder()

        query.append("ALTER WAREHOUSE {full_name:i} SET", {
            "full_name": bp.full_name
        })

        query.append(self._build_common_properties(bp))

        self.engine.execute_safe_ddl(query)

        return True

    def _compare_resource_monitor(self, bp: WarehouseBlueprint, row: dict):
        if bp.resource_monitor is None and row['resource_monitor']:
            self.engine.execute_safe_ddl("ALTER WAREHOUSE {full_name:i} UNSET RESOURCE_MONITOR", {
                "full_name": bp.full_name,
            }, condition=self.engine.settings.execute_resource_monitor)

            return True

        if bp.resource_monitor and str(bp.resource_monitor) != row['resource_monitor']:
            self.engine.execute_safe_ddl("ALTER WAREHOUSE {full_name:i} SET RESOURCE_MONITOR = {resource_monitor:i}", {
                "full_name": bp.full_name,
                "resource_monitor": bp.resource_monitor,
            }, condition=self.engine.settings.execute_resource_monitor)

            return True

        return False

    def _build_common_properties(self, bp: WarehouseBlueprint):
        query = self.engine.query_builder()

        query.append_nl("WAREHOUSE_SIZE = {size}", {
            "size": self._normalise_warehouse_size(bp.size),
        })

        if bp.min_cluster_count:
            query.append_nl("MIN_CLUSTER_COUNT = {min_cluster_count:d}", {
                "min_cluster_count": bp.min_cluster_count,
            })

        if bp.max_cluster_count:
            query.append_nl("MAX_CLUSTER_COUNT = {max_cluster_count:d}", {
                "max_cluster_count": bp.max_cluster_count,
            })

        if bp.scaling_policy:
            query.append_nl("SCALING_POLICY = {scaling_policy}", {
                "scaling_policy": bp.scaling_policy,
            })

        query.append_nl("AUTO_SUSPEND = {auto_suspend:d}", {
            "auto_suspend": bp.auto_suspend,
        })

        return query

    def _normalise_warehouse_size(self, size: str):
        return size.upper().replace('-', '')

    def _post_process(self):
        if not self.engine.context.current_warehouse:
            return

        for result in self.resolved_objects.values():
            if result == ResolveResult.CREATE:
                # Revert current warehouse to original state if at least one object was created
                self.engine.execute_context_ddl("USE WAREHOUSE {full_name:i}", {
                    "full_name": self.engine.context.current_warehouse
                })

                break



from abc import abstractmethod

from snowddl.blueprint import RoleBlueprint, Grant, FutureGrant, ComplexIdentWithPrefix, ComplexIdentWithPrefixAndArgs, BaseDataType
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class AbstractRoleResolver(AbstractResolver):
    @abstractmethod
    def get_role_suffix(self) -> str:
        pass

    def get_object_type(self):
        return ObjectType.ROLE

    def get_existing_objects(self):
        existing_roles = {}

        cur = self.engine.execute_meta("SHOW ROLES LIKE {pattern:lse}", {
            'pattern': (self.config.env_prefix, self.get_role_suffix()),
        })

        for r in cur:
            if r['owner'] != self.engine.context.current_role:
                continue

            existing_roles[r['name']] = {
                "role_name": r['name'],
                "grants": self.get_existing_grants(r['name']),
                "future_grants": self.get_existing_future_grants(r['name']),
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_roles

    def get_existing_grants(self, role_name):
        grants = []

        cur = self.engine.execute_meta("SHOW GRANTS TO ROLE {role_name:i}", {
            "role_name": role_name,
        })

        for r in cur:
            grants.append(Grant(
                privilege=r['privilege'],
                on=ObjectType[r['granted_on']],
                name=self.build_grant_name_ident(r['name']),
            ))

        return grants

    def get_existing_future_grants(self, role_name):
        future_grants = []

        cur = self.engine.execute_meta("SHOW FUTURE GRANTS TO ROLE {role_name:i}", {
            "role_name": role_name,
        })

        for r in cur:
            future_grants.append(FutureGrant(
                privilege=r['privilege'],
                on=ObjectType[r['grant_on']],
                name=self.build_grant_name_ident(r['name']),
            ))

        return future_grants

    def build_grant_name_ident(self, grant_name):
        parts = [p.strip('"') for p in grant_name.split('.')]
        last_part = parts[-1]

        # Remove object type component from future grant names
        if last_part.startswith('<') and last_part.endswith('>'):
            parts.pop()
        # Extract data types for arguments of functions and procedures
        elif '(' in last_part:
            start_dtypes_idx = last_part.index('(')
            finish_dtypes_idx = last_part.index(')')

            parts[-1] = last_part[0:start_dtypes_idx]

            arguments = last_part[start_dtypes_idx+1:finish_dtypes_idx]
            data_types = [BaseDataType[arg.strip(' ').split(' ')[1]] for arg in arguments.split(',')] if arguments else []

            return ComplexIdentWithPrefixAndArgs('', *parts, data_types=data_types)

        return ComplexIdentWithPrefix('', *parts)

    def create_object(self, bp: RoleBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE ROLE {role_name:i}", {
            "role_name": bp.full_name,
        })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_safe_ddl(query)

        self.engine.execute_safe_ddl("GRANT ROLE {role_name:i} TO ROLE {current_role:i}", {
            "role_name": bp.full_name,
            "current_role": self.engine.context.current_role,
        })

        for bp_grant in bp.grants:
            self.create_grant(bp.full_name, bp_grant)

        for bp_future_grant in bp.future_grants:
            self.create_future_grant(bp.full_name, bp_future_grant)

        return ResolveResult.CREATE

    def compare_object(self, bp: RoleBlueprint, row: dict):
        result = ResolveResult.NOCHANGE

        if bp.comment != row['comment']:
            self.engine.execute_safe_ddl("ALTER ROLE {role_name:i} SET COMMENT = {comment}", {
                "role_name": bp.full_name,
                "comment": bp.comment,
            })

            result = ResolveResult.ALTER

        for bp_grant in bp.grants:
            if bp_grant not in row['grants']:
                self.create_grant(bp.full_name, bp_grant)
                result = ResolveResult.GRANT

        for existing_grant in row['grants']:
            if existing_grant not in bp.grants \
            and existing_grant.on.is_future_grant_supported \
            and self.grant_to_future_grant(existing_grant) not in bp.future_grants:
                self.drop_grant(bp.full_name, existing_grant)
                result = ResolveResult.GRANT

        for bp_future_grant in bp.future_grants:
            if bp_future_grant not in row['future_grants']:
                self.create_future_grant(bp.full_name, bp_future_grant)
                result = ResolveResult.GRANT

            if self.engine.settings.refresh_future_grants:
                self.refresh_future_grant(bp.full_name, bp_future_grant)
                result = ResolveResult.GRANT

        for existing_future_grant in row['future_grants']:
            if existing_future_grant not in bp.future_grants:
                self.drop_future_grant(bp.full_name, existing_future_grant)
                result = ResolveResult.GRANT

        return result

    def drop_object(self, row: dict):
        # All grants and future grants are destroyed automatically
        self.engine.execute_safe_ddl("DROP ROLE {role_name:i}", {
            "role_name": row['role_name'],
        })

        return ResolveResult.DROP

    def create_grant(self, role_name, grant: Grant):
        if grant.privilege == "USAGE" and grant.on == ObjectType.ROLE:
            self.engine.execute_safe_ddl("GRANT ROLE {name:i} TO ROLE {role_name:i}", {
                "name": grant.name,
                "role_name": role_name,
            })
        else:
            self.engine.execute_safe_ddl("GRANT {privilege:r} ON {on:r} {name:i} TO ROLE {role_name:i}", {
                "privilege": grant.privilege,
                "on": grant.on.singular,
                "name": grant.name,
                "role_name": role_name,
            })

    def drop_grant(self, role_name, grant: Grant):
        if grant.privilege == "USAGE" and grant.on == ObjectType.ROLE:
            self.engine.execute_safe_ddl("REVOKE ROLE {name:i} FROM ROLE {role_name:i}", {
                "name": grant.name,
                "role_name": role_name,
            })
        else:
            self.engine.execute_safe_ddl("REVOKE {privilege:r} ON {on:r} {name:i} FROM ROLE {role_name:i}", {
                "privilege": grant.privilege,
                "on": grant.on.singular,
                "name": grant.name,
                "role_name": role_name,
            })

    def create_future_grant(self, role_name, grant: FutureGrant):
        self.engine.execute_safe_ddl("GRANT {privilege:r} ON FUTURE {on_plural:r} IN SCHEMA {name:i} TO ROLE {role_name:i}", {
            "privilege": grant.privilege,
            "on_plural": grant.on.plural,
            "name": grant.name,
            "role_name": role_name,
        })

    def drop_future_grant(self, role_name, grant: FutureGrant):
        self.engine.execute_safe_ddl("REVOKE {privilege:r} ON FUTURE {on_plural:r} IN SCHEMA {name:i} FROM ROLE {role_name:i}", {
            "privilege": grant.privilege,
            "on_plural": grant.on.plural,
            "name": grant.name,
            "role_name": role_name,
        })

    def refresh_future_grant(self, role_name, grant: FutureGrant):
        # Bulk grant on objects of type PIPE to ROLE is restricted (by Snowflake)
        if grant.on == ObjectType.PIPE:
            return

        self.engine.execute_safe_ddl("GRANT {privilege:r} ON ALL {on_plural:r} IN SCHEMA {name:i} TO ROLE {role_name:i}", {
            "privilege": grant.privilege,
            "on_plural": grant.on.plural,
            "name": grant.name,
            "role_name": role_name,
        })

    def grant_to_future_grant(self, grant: Grant):
        future_grant = FutureGrant(
            privilege=grant.privilege,
            on=grant.on,
            name=ComplexIdentWithPrefix(grant.name.env_prefix, *grant.name.parts[:-1]) if isinstance(grant.name, ComplexIdentWithPrefix) else grant.name
        )

        return future_grant



from snowddl.blueprint import RoleBlueprint, WarehouseBlueprint, Grant
from snowddl.resolver.abc_role_resolver import AbstractRoleResolver, ObjectType


class WarehouseRoleResolver(AbstractRoleResolver):
    def get_role_suffix(self):
        return self.config.WAREHOUSE_ROLE_SUFFIX

    def get_blueprints(self):
        blueprints = []

        for warehouse in self.config.get_blueprints_by_type(WarehouseBlueprint).values():
            blueprints.append(self.get_blueprint_usage_role(warehouse))
            blueprints.append(self.get_blueprint_monitor_role(warehouse))

        return {str(bp.full_name): bp for bp in blueprints}

    def get_blueprint_usage_role(self, warehouse: WarehouseBlueprint):
        grants = []

        grants.append(Grant(
            privilege="USAGE",
            on=ObjectType.WAREHOUSE,
            name=warehouse.full_name,
        ))

        grants.append(Grant(
            privilege="OPERATE",
            on=ObjectType.WAREHOUSE,
            name=warehouse.full_name,
        ))

        bp = RoleBlueprint(
            full_name=self.config.build_role_ident(warehouse.full_name, 'USAGE', self.get_role_suffix()),
            grants=grants,
            future_grants=[],
            comment=None,
        )

        return bp

    def get_blueprint_monitor_role(self, warehouse: WarehouseBlueprint):
        grants = []

        grants.append(Grant(
            privilege="MONITOR",
            on=ObjectType.WAREHOUSE,
            name=warehouse.full_name,
        ))

        grants.append(Grant(
            privilege="OPERATE",
            on=ObjectType.WAREHOUSE,
            name=warehouse.full_name,
        ))

        bp = RoleBlueprint(
            full_name=self.config.build_role_ident(warehouse.full_name, 'MONITOR', self.get_role_suffix()),
            grants=grants,
            future_grants=[],
            comment=None,
        )

        return bp



from snowddl.blueprint import StageBlueprint, StageFileBlueprint, Ident, IdentWithPrefix, ComplexIdentWithPrefix, ComplexIdentWithPrefixAndPath
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


stage_json_schema = {
    "type": "object",
    "properties": {
        "url": {
            "type": "string"
        },
        "storage_integration": {
            "type": "string"
        },
        "encryption": {
            "type": "object",
            "additionalProperties": {
                "type": ["boolean", "number", "string"]
            }
        },
        "directory": {
            "type": "object",
            "additionalProperties": {
                "type": ["boolean", "number", "string"]
            }
        },
        "file_format": {
            "type": "string"
        },
        "copy_options": {
            "type": "object",
            "additionalProperties": {
                "type": ["array", "boolean", "number", "string"]
            }
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}


class StageParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("stage", stage_json_schema, self.process_stage)

    def process_stage(self, f: ParsedFile):
        stage_files_dir = (f.path.parent / f.name)

        if stage_files_dir.is_dir() and f.params.get('url'):
            raise ValueError("External stage cannot have managed stage files")

        # Stage
        bp = StageBlueprint(
            full_name=ComplexIdentWithPrefix(self.env_prefix, f.database, f.schema, f.name),
            database=IdentWithPrefix(self.env_prefix, f.database),
            schema=Ident(f.schema),
            name=Ident(f.name),
            url=f.params.get('url'),
            storage_integration=Ident(f.params['storage_integration']) if f.params.get('storage_integration') else None,
            encryption=self.normalise_params_dict(f.params.get('encryption')),
            directory=self.normalise_params_dict(f.params.get('directory')),
            file_format=self.config.build_complex_ident(f.params['file_format'], f.database, f.schema) if f.params.get('file_format') else None,
            copy_options=self.normalise_params_dict(f.params.get('copy_options')),
            upload_stage_files=stage_files_dir.is_dir(),
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)

        # Stage files
        if stage_files_dir.is_dir():
            for path in stage_files_dir.glob('**/*'):
                if not path.is_file():
                    continue

                stage_path = f"/{path.relative_to(stage_files_dir)}"

                bp = StageFileBlueprint(
                    full_name=ComplexIdentWithPrefixAndPath(self.env_prefix, f.database, f.schema, f.name, path=stage_path),
                    local_path=str(path),
                    stage_name=ComplexIdentWithPrefix(self.env_prefix, f.database, f.schema, f.name),
                    stage_path=stage_path,
                    comment=None
                )

                self.config.add_blueprint(bp)



from dataclasses import dataclass
from typing import List, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from .ident import Ident, ComplexIdentWithPrefix
    from .object_type import ObjectType


@dataclass
class MaskingPolicyReference:
    object_type: "ObjectType"
    object_name: "ComplexIdentWithPrefix"
    columns: List["Ident"]


@dataclass
class RowAccessPolicyReference:
    object_type: "ObjectType"
    object_name: "ComplexIdentWithPrefix"
    columns: List["Ident"]


@dataclass
class TagReference:
    object_type: "ObjectType"
    object_name: "ComplexIdentWithPrefix"
    column_name: Optional["Ident"]
    tag_value: str



from dataclasses import dataclass
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .ident import Ident
    from .data_type import DataType


@dataclass
class ExternalTableColumn:
    name: "Ident"
    type: "DataType"
    expr: str
    not_null: bool
    comment: Optional[str]


@dataclass
class TableColumn:
    name: "Ident"
    type: "DataType"
    not_null: bool
    default: Optional[str]
    comment: Optional[str]


@dataclass
class ViewColumn:
    name: "Ident"
    comment: Optional[str]


@dataclass
class NameWithType:
    name: "Ident"
    type: "DataType"



from json import loads

from snowddl.blueprint import RowAccessPolicyBlueprint, ObjectType, Edition
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult


class RowAccessPolicyResolver(AbstractSchemaObjectResolver):
    skip_on_empty_blueprints = True
    skip_min_edition = Edition.ENTERPRISE

    def get_object_type(self) -> ObjectType:
        return ObjectType.ROW_ACCESS_POLICY

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW ROW ACCESS POLICIES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"

            existing_objects[full_name] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(RowAccessPolicyBlueprint)

    def create_object(self, bp: RowAccessPolicyBlueprint):
        self._create_policy(bp)
        self._apply_policy_refs(bp, skip_existing=True)

        return ResolveResult.CREATE

    def compare_object(self, bp: RowAccessPolicyBlueprint, row: dict):
        cur = self.engine.execute_meta("DESC ROW ACCESS POLICY {full_name:i}", {
            "full_name": bp.full_name,
        })

        r = cur.fetchone()

        # If signature was changed, policy and all references must be dropped and created again
        if r['signature'] != f"({', '.join([f'{a.name} {a.type.base_type.name}' for a in bp.arguments])})":
            self._drop_policy_refs(bp.database, bp.schema, bp.name)
            self._drop_policy(bp.database, bp.schema, bp.name)

            self._create_policy(bp)
            self._apply_policy_refs(bp, skip_existing=True)

            return ResolveResult.REPLACE

        result = ResolveResult.NOCHANGE

        if self._apply_policy_refs(bp):
            result = ResolveResult.ALTER

        if r['body'] != bp.body:
            self.engine.execute_unsafe_ddl("ALTER ROW ACCESS POLICY {full_name:i} SET BODY -> {body:r}", {
                "full_name": bp.full_name,
                "body": bp.body,
            }, condition=self.engine.settings.execute_row_access_policy)

            result = ResolveResult.ALTER

        if row['comment'] != bp.comment:
            self.engine.execute_unsafe_ddl("ALTER ROW ACCESS POLICY {full_name:i} SET COMMENT = {comment}", {
                "full_name": bp.full_name,
                "comment": bp.comment,
            }, condition=self.engine.settings.execute_row_access_policy)

            result = ResolveResult.ALTER

        return result

    def drop_object(self, row: dict):
        self._drop_policy_refs(row['database'], row['schema'], row['name'])
        self._drop_policy(row['database'], row['schema'], row['name'])

        return ResolveResult.DROP

    def _create_policy(self, bp: RowAccessPolicyBlueprint):
        query = self.engine.query_builder()

        query.append("CREATE ROW ACCESS POLICY {full_name:i} AS (", {
            "full_name": bp.full_name,
        })

        for idx, arg in enumerate(bp.arguments):
            query.append_nl("    {comma:r}{arg_name:i} {arg_type:r}", {
                "comma": "  " if idx == 0 else ", ",
                "arg_name": arg.name,
                "arg_type": arg.type,
            })

        query.append_nl(")")

        query.append_nl("RETURNS BOOLEAN -> ")

        query.append_nl("{body:r}", {
            "body": bp.body,
        })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        self.engine.execute_unsafe_ddl(query, condition=self.engine.settings.execute_row_access_policy)

    def _drop_policy(self, database, schema, name):
        self.engine.execute_unsafe_ddl("DROP ROW ACCESS POLICY {database:i}.{schema:i}.{name:i}", {
            "database": database,
            "schema": schema,
            "name": name,
        }, condition=self.engine.settings.execute_row_access_policy)

    def _apply_policy_refs(self, bp: RowAccessPolicyBlueprint, skip_existing=False):
        existing_policy_refs = {} if skip_existing else self._get_existing_policy_refs(bp.database, bp.schema, bp.name)
        applied_change = False

        for ref in bp.references:
            ref_key = f"{ref.object_type.name}|{ref.object_name}|{bp.full_name}"

            # Policy was applied before
            if ref_key in existing_policy_refs:
                existing_ref_columns = existing_policy_refs[ref_key]['columns']
                del(existing_policy_refs[ref_key])

                # Policy was applied to exactly the same columns
                if [str(c) for c in ref.columns] == existing_ref_columns:
                    continue

                # Policy was applied to different columns, should be dropped and re-created
                self.engine.execute_unsafe_ddl("ALTER {object_type:r} {object_name:i} DROP ROW ACCESS POLICY {policy_name:i}", {
                    "object_type": ref.object_type.simplified,
                    "object_name": ref.object_name,
                    "policy_name": bp.full_name,
                }, condition=self.engine.settings.execute_row_access_policy)

            # Apply new policy
            self.engine.execute_unsafe_ddl("ALTER {object_type:r} {object_name:i} ADD ROW ACCESS POLICY {policy_name:i} ON ({columns:i})", {
                "object_type": ref.object_type.simplified,
                "object_name": ref.object_name,
                "policy_name": bp.full_name,
                "columns": ref.columns,
            }, condition=self.engine.settings.execute_row_access_policy)

            applied_change = True

        # Remove remaining policy references which no longer exist in blueprint
        for existing_ref in existing_policy_refs.values():
            self.engine.execute_unsafe_ddl("ALTER {object_type:r} {database:i}.{schema:i}.{name:i} DROP ROW ACCESS POLICY {policy_name:i}", {
                "object_type": existing_ref['object_type'],
                "database": existing_ref['database'],
                "schema": existing_ref['schema'],
                "name": existing_ref['name'],
                "policy_name": bp.full_name,
            }, condition=self.engine.settings.execute_row_access_policy)

            applied_change = True

        return applied_change

    def _drop_policy_refs(self, database, schema, name):
        existing_policy_refs = self._get_existing_policy_refs(database, schema, name)

        for existing_ref in existing_policy_refs.values():
            self.engine.execute_unsafe_ddl("ALTER {object_type:r} {database:i}.{schema:i}.{name:i} DROP ROW ACCESS POLICY {policy_database:i}.{policy_schema:i}.{policy_name:i}", {
                "object_type": existing_ref['object_type'],
                "database": existing_ref['database'],
                "schema": existing_ref['schema'],
                "name": existing_ref['name'],
                "policy_database": database,
                "policy_schema": schema,
                "policy_name": name,
            }, condition=self.engine.settings.execute_row_access_policy)

    def _get_existing_policy_refs(self, database, schema, name):
        existing_policy_refs = {}

        cur = self.engine.execute_meta("SELECT * FROM TABLE({database:i}.information_schema.policy_references(policy_name => {policy_name}))", {
            "database": database,
            "policy_name": f"{database}.{schema}.{name}"
        })

        for r in cur:
            ref_key = f"{r['REF_ENTITY_DOMAIN']}" \
                      f"|{r['REF_DATABASE_NAME']}.{r['REF_SCHEMA_NAME']}.{r['REF_ENTITY_NAME']}" \
                      f"|{r['POLICY_DB']}.{r['POLICY_SCHEMA']}.{r['POLICY_NAME']}"

            existing_policy_refs[ref_key] = {
                "object_type": r['REF_ENTITY_DOMAIN'],
                "database": r['REF_DATABASE_NAME'],
                "schema": r['REF_SCHEMA_NAME'],
                "name": r['REF_ENTITY_NAME'],
                "columns": loads(r['REF_ARG_COLUMN_NAMES']),
            }

        return existing_policy_refs



from snowddl.blueprint import WarehouseBlueprint, Ident, IdentWithPrefix
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


warehouse_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": "object",
        "properties": {
            "size": {
                "type": "string"
            },
            "min_cluster_count": {
                "type": "integer"
            },
            "max_cluster_count": {
                "type": "integer"
            },
            "scaling_policy": {
                "type": "string"
            },
            "auto_suspend": {
                "type": "integer"
            },
            "resource_monitor": {
                "type": "string"
            },
            "comment": {
                "type": "string"
            }
        },
        "required": ["size"],
        "additionalProperties": False
    }
}


class WarehouseParser(AbstractParser):
    def load_blueprints(self):
        self.parse_single_file(self.base_path / 'warehouse.yaml', warehouse_json_schema, self.process_warehouse)

    def process_warehouse(self, f: ParsedFile):
        for warehouse_name, warehouse in f.params.items():
            bp = WarehouseBlueprint(
                full_name=IdentWithPrefix(self.env_prefix, warehouse_name),
                size=warehouse['size'],
                auto_suspend=warehouse.get('auto_suspend', 60),
                min_cluster_count=warehouse.get('min_cluster_count'),
                max_cluster_count=warehouse.get('max_cluster_count'),
                scaling_policy=warehouse.get('scaling_policy'),
                resource_monitor=Ident(warehouse['resource_monitor']) if warehouse.get('resource_monitor') else None,
                comment=warehouse.get('comment'),
            )

            self.config.add_blueprint(bp)



from snowddl.blueprint import SchemaBlueprint, Ident, IdentWithPrefix, ComplexIdentWithPrefix, Grant, ObjectType
from snowddl.parser.abc_parser import AbstractParser, ParsedFile
from snowddl.parser.database import database_json_schema


schema_json_schema = {
    "type": "object",
    "properties": {
        "is_transient": {
            "type": "boolean"
        },
        "retention_time": {
            "type": "integer"
        },
        "is_sandbox": {
            "type": "boolean"
        },
        "owner_schema_read": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "owner_schema_write": {
            "type": "array",
            "items": {
                "type": "string"
            }
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}


class SchemaParser(AbstractParser):
    def load_blueprints(self):
        for database_path in self.base_path.iterdir():
            if not database_path.is_dir():
                continue

            database_params = self.parse_single_file(database_path / 'params.yaml', database_json_schema)

            for schema_path in database_path.iterdir():
                if not schema_path.is_dir():
                    continue

                schema_params = self.parse_single_file(schema_path / 'params.yaml', schema_json_schema)

                combined_params = {
                    "is_transient": database_params.get('is_transient', False) or schema_params.get('is_transient', False),
                    "retention_time": schema_params.get("retention_time"),
                    "is_sandbox": database_params.get('is_sandbox', False) or schema_params.get('is_sandbox', False),
                }

                owner_additional_grants = []

                for full_schema_name in schema_params.get('owner_schema_read', []):
                    owner_additional_grants.append(self.build_schema_role_grant(full_schema_name, 'READ'))

                for full_schema_name in schema_params.get('owner_schema_write', []):
                    owner_additional_grants.append(self.build_schema_role_grant(full_schema_name, 'WRITE'))

                bp = SchemaBlueprint(
                    full_name=ComplexIdentWithPrefix(self.env_prefix, database_path.name, schema_path.name),
                    database=IdentWithPrefix(self.env_prefix, database_path.name),
                    schema=Ident(schema_path.name),
                    is_transient=combined_params.get('is_transient', False),
                    retention_time=combined_params.get('retention_time', None),
                    is_sandbox=combined_params.get('is_sandbox', False),
                    owner_additional_grants=owner_additional_grants,
                    comment=schema_params.get('comment', None),
                )

                self.config.add_blueprint(bp)

    def build_schema_role_grant(self, full_schema_name, grant_type):
        database, schema = full_schema_name.split('.')

        return Grant(
            privilege="USAGE",
            on=ObjectType.ROLE,
            name=self.config.build_role_ident(database, schema, grant_type, self.config.SCHEMA_ROLE_SUFFIX),
        )



from pathlib import Path

from snowddl.config import SnowDDLConfig
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


placeholder_json_schema = {
    "type": "object",
    "additionalProperties": {
        "type": ["boolean", "number", "string"]
    }
}


class PlaceholderParser(AbstractParser):
    def __init__(self, config: SnowDDLConfig, base_path: Path, placeholder_path: Path = None):
        super().__init__(config, base_path)
        self.placeholder_path = placeholder_path

    def load_blueprints(self):
        # ENV_PREFIX placeholder is added automatically and always available
        self.config.add_placeholder('ENV_PREFIX', self.env_prefix)

        # Add other placeholders from file
        self.parse_single_file(self.base_path / 'placeholder.yaml', placeholder_json_schema, self.process_placeholder)

    def process_placeholder(self, f: ParsedFile):
        placeholders = f.params

        if self.placeholder_path:
            overloaded_placeholders = self.parse_single_file(self.placeholder_path, placeholder_json_schema)
            placeholders.update(overloaded_placeholders)

        for name, value in placeholders.items():
            name = str(name).upper()
            self.config.add_placeholder(name, value)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from snowddl.engine import SnowDDLEngine


class SnowDDLSchemaCache:
    def __init__(self, engine: "SnowDDLEngine"):
        self.engine = engine

        self.databases = {}
        self.schemas = {}

        self.reload()

    def reload(self):
        self.databases = {}
        self.schemas = {}

        cur = self.engine.execute_meta("SHOW DATABASES LIKE {env_prefix:ls}", {
            'env_prefix': self.engine.config.env_prefix,
        })

        for r in cur:
            # Skip databases created by other roles
            if r['owner'] != self.engine.context.current_role:
                continue

            # Skip shares
            if r['origin']:
                continue

            self.databases[r['name']] = {
                "database": r['name'],
                "owner": r['owner'],
                "comment": r['comment'] if r['comment'] else None,
                "is_transient": "TRANSIENT" in r['options'],
                "retention_time": int(r['retention_time']),
            }

        # Process schemas in parallel
        for database_schemas in self.engine.executor.map(self._get_database_schemas, self.databases):
            self.schemas.update(database_schemas)

    def _get_database_schemas(self, database_name):
        schemas = {}

        cur = self.engine.execute_meta("SHOW SCHEMAS IN DATABASE {database:i}", {
            "database": database_name,
        })

        for r in cur:
            # Skip schemas created by other roles
            if r['owner'] != self.engine.context.current_role:
                continue

            # Skip INFORMATION_SCHEMA
            if r['name'] == "INFORMATION_SCHEMA":
                continue

            schemas[f"{r['database_name']}.{r['name']}"] = {
                "database": r['database_name'],
                "schema": r['name'],
                "owner": r['owner'],
                "comment": r['comment'] if r['comment'] else None,
                "is_transient": "TRANSIENT" in r['options'],
                "is_managed_access": "MANAGED ACCESS" in r['options'],
                "retention_time": int(r['retention_time']) if r['retention_time'].isdigit() else 0,
            }

        return schemas



from abc import ABC
from dataclasses import dataclass
from typing import Optional, List, Dict, Union, TypeVar, TYPE_CHECKING

from .column import ExternalTableColumn, TableColumn, ViewColumn, NameWithType
from .data_type import BaseDataType, DataType
from .grant import Grant, FutureGrant
from .ident import Ident, IdentWithPrefix, ComplexIdentWithPrefix, ComplexIdentWithPrefixAndArgs, ComplexIdentWithPrefixAndPath
from .reference import MaskingPolicyReference, RowAccessPolicyReference, TagReference
from .stage import StageWithPath

if TYPE_CHECKING:
    from .object_type import ObjectType


@dataclass
class AbstractBlueprint(ABC):
    full_name: Ident
    comment: Optional[str]


@dataclass
class SchemaObjectBlueprint(AbstractBlueprint, ABC):
    full_name: ComplexIdentWithPrefix
    database: IdentWithPrefix
    schema: Ident
    name: Ident


@dataclass
class RoleBlueprint(AbstractBlueprint):
    full_name: IdentWithPrefix
    grants: List[Grant]
    future_grants: List[FutureGrant]


@dataclass
class DependsOnMixin(ABC):
    full_name: Ident
    depends_on: List[Ident]


@dataclass
class AccountParameterBlueprint(AbstractBlueprint):
    value: Union[bool,float,int,str]


@dataclass
class BusinessRoleBlueprint(RoleBlueprint):
    pass


@dataclass
class DatabaseBlueprint(AbstractBlueprint):
    full_name: IdentWithPrefix
    database: IdentWithPrefix
    is_transient: Optional[bool]
    retention_time: Optional[int]
    is_sandbox: Optional[bool]


@dataclass
class ExternalFunctionBlueprint(SchemaObjectBlueprint):
    full_name: ComplexIdentWithPrefixAndArgs
    arguments: List[NameWithType]
    returns: DataType
    api_integraton: Ident
    url: str
    is_secure: bool
    is_strict: bool
    is_immutable: bool
    headers: Optional[Dict[str,str]]
    context_headers: Optional[List[Ident]]
    max_batch_rows: Optional[int]
    compression: Optional[str]
    request_translator: Optional[ComplexIdentWithPrefix]
    response_translator: Optional[ComplexIdentWithPrefix]


@dataclass
class ExternalTableBlueprint(SchemaObjectBlueprint):
    columns: Optional[List[ExternalTableColumn]]
    partition_by: Optional[List[Ident]]
    location_stage: ComplexIdentWithPrefix
    location_path: Optional[str]
    location_pattern: Optional[str]
    file_format: ComplexIdentWithPrefix
    refresh_on_create: Optional[bool]
    auto_refresh: Optional[bool]
    aws_sns_topic: Optional[str]
    integration: Optional[Ident]


@dataclass
class FileFormatBlueprint(SchemaObjectBlueprint):
    type: str
    format_options: Optional[Dict[str,Union[bool,float,int,str,list]]]


@dataclass
class ForeignKeyBlueprint(AbstractBlueprint):
    table_name: ComplexIdentWithPrefix
    columns: List[Ident]
    ref_table_name: ComplexIdentWithPrefix
    ref_columns: List[Ident]


@dataclass
class FunctionBlueprint(SchemaObjectBlueprint):
    full_name: ComplexIdentWithPrefixAndArgs
    language: str
    body: Optional[str]
    arguments: List[NameWithType]
    returns: Union[DataType,List[NameWithType]]
    is_secure: bool
    is_strict: bool
    is_immutable: bool
    runtime_version: Optional[str]
    imports: Optional[List[StageWithPath]]
    handler: Optional[str]


@dataclass
class MaterializedViewBlueprint(SchemaObjectBlueprint):
    text: str
    columns: Optional[List[ViewColumn]]
    is_secure: Optional[bool]
    cluster_by: Optional[List[str]]


@dataclass
class MaskingPolicyBlueprint(SchemaObjectBlueprint):
    full_name: ComplexIdentWithPrefix
    arguments: List[NameWithType]
    returns: DataType
    body: str
    references: List[MaskingPolicyReference]


@dataclass
class NetworkPolicyBlueprint(AbstractBlueprint):
    full_name: Ident
    allowed_ip_list: List[str]
    blocked_ip_list: List[str]


@dataclass
class PipeBlueprint(SchemaObjectBlueprint):
    auto_ingest: bool
    copy_table_name: ComplexIdentWithPrefix
    copy_stage_name: ComplexIdentWithPrefix
    copy_path: Optional[str]
    copy_pattern: Optional[str]
    copy_transform: Optional[Dict[str, str]]
    copy_file_format: Optional[ComplexIdentWithPrefix]
    copy_options: Optional[Dict[str,Union[bool,float,int,str,list]]]
    aws_sns_topic: Optional[str]
    integration: Optional[Ident]


@dataclass
class PrimaryKeyBlueprint(AbstractBlueprint):
    table_name: ComplexIdentWithPrefix
    columns: List[Ident]


@dataclass
class ProcedureBlueprint(SchemaObjectBlueprint):
    full_name: ComplexIdentWithPrefixAndArgs
    language: str
    body: str
    arguments: List[NameWithType]
    returns: DataType
    is_strict: bool
    is_immutable: bool
    is_execute_as_caller: bool


@dataclass
class ResourceMonitorBlueprint(AbstractBlueprint):
    full_name: Ident
    credit_quota: int
    frequency: str
    triggers: Dict[int,str]


@dataclass
class RowAccessPolicyBlueprint(SchemaObjectBlueprint):
    full_name: ComplexIdentWithPrefix
    arguments: List[NameWithType]
    body: str
    references: List[RowAccessPolicyReference]


@dataclass
class SchemaBlueprint(AbstractBlueprint):
    full_name: ComplexIdentWithPrefix
    database: IdentWithPrefix
    schema: Ident
    is_transient: Optional[bool]
    retention_time: Optional[int]
    is_sandbox: Optional[bool]
    owner_additional_grants: List[Grant]


@dataclass
class SchemaRoleBlueprint(RoleBlueprint, DependsOnMixin):
    pass


@dataclass
class StageBlueprint(SchemaObjectBlueprint):
    url: Optional[str]
    storage_integration: Optional[Ident]
    encryption: Optional[Dict[str,str]]
    directory: Optional[Dict[str,str]]
    file_format: Optional[ComplexIdentWithPrefix]
    copy_options: Optional[Dict[str,Union[bool,float,int,str,list]]]
    upload_stage_files: bool


@dataclass
class StageFileBlueprint(AbstractBlueprint):
    full_name: ComplexIdentWithPrefixAndPath
    local_path: str
    stage_name: ComplexIdentWithPrefix
    stage_path: str


@dataclass
class SequenceBlueprint(SchemaObjectBlueprint):
    start: int
    interval: int


@dataclass
class StreamBlueprint(SchemaObjectBlueprint):
    object_type: "ObjectType"
    object_name: ComplexIdentWithPrefix
    append_only: Optional[bool]
    insert_only: Optional[bool]
    show_initial_rows: Optional[bool]


@dataclass
class TableBlueprint(SchemaObjectBlueprint):
    columns: List[TableColumn]
    cluster_by: Optional[List[str]]
    is_transient: bool
    retention_time: Optional[int]
    change_tracking: bool
    search_optimization: bool


@dataclass
class TagBlueprint(SchemaObjectBlueprint):
    full_name: ComplexIdentWithPrefix
    references: List[TagReference]


@dataclass
class TaskBlueprint(SchemaObjectBlueprint, DependsOnMixin):
    body: str
    schedule: Optional[str]
    after: Optional[Ident]
    when: Optional[str]
    warehouse: Optional[IdentWithPrefix]
    user_task_managed_initial_warehouse_size: Optional[str]
    allow_overlapping_execution: Optional[bool]
    session_params: Optional[Dict[str,Union[bool,float,int,str]]]
    user_task_timeout_ms: Optional[int]


@dataclass
class TechRoleBlueprint(RoleBlueprint):
    pass


@dataclass
class UniqueKeyBlueprint(AbstractBlueprint):
    table_name: ComplexIdentWithPrefix
    columns: List[Ident]


@dataclass
class UserBlueprint(AbstractBlueprint):
    full_name: IdentWithPrefix
    first_name: Optional[str]
    last_name: Optional[str]
    email: Optional[str]
    disabled: bool
    business_roles: List[IdentWithPrefix]
    password: Optional[str]
    rsa_public_key: Optional[str]
    rsa_public_key_2: Optional[str]
    default_warehouse: Optional[ComplexIdentWithPrefix]
    default_namespace: Optional[ComplexIdentWithPrefix]
    session_params: Optional[Dict[str,Union[bool,float,int,str]]]


@dataclass
class ViewBlueprint(SchemaObjectBlueprint, DependsOnMixin):
    text: str
    columns: Optional[List[ViewColumn]]
    is_secure: Optional[bool]


@dataclass
class WarehouseBlueprint(AbstractBlueprint):
    full_name: IdentWithPrefix
    size: str
    auto_suspend: int
    min_cluster_count: Optional[int]
    max_cluster_count: Optional[int]
    scaling_policy: Optional[str]
    resource_monitor: Optional[Ident]


T_Blueprint = TypeVar('T_Blueprint', bound=AbstractBlueprint)



from string import ascii_letters, digits
from typing import List

from .data_type import BaseDataType


class Ident:
    allowed_chars = set(ascii_letters + digits + '_')

    def __init__(self, value):
        self.value = self.validate_ident(value)

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"<{self.__class__.__name__}={str(self)}>"

    def __eq__(self, other):
        if not isinstance(other, Ident):
            raise NotImplemented

        return str(self) == str(other)

    def validate_ident(self, value):
        if isinstance(value, Ident):
            value = str(value.value)
        else:
            value = str(value)

        if not value:
            ValueError("Identifier cannot be empty")

        for char in value:
            if char not in self.allowed_chars:
                ValueError(f"Character [{char}] in not allowed in identifier [{value}], only ASCII letters, digits and single underscores are accepted")

        return value.upper()


class IdentWithPrefix(Ident):
    def __init__(self, env_prefix, value):
        self.env_prefix = env_prefix
        self.value = self.validate_ident(value)

    def __str__(self):
        return f"{self.env_prefix}{self.value}"


class ComplexIdentWithPrefix(IdentWithPrefix):
    def __init__(self, env_prefix, *parts):
        self.env_prefix = env_prefix
        self.parts = tuple(self.validate_ident(part) for part in parts)

    def parts_for_format(self):
        parts_for_format = []

        for idx, part in enumerate(self.parts):
            if idx == 0:
                parts_for_format.append(f"{self.env_prefix}{part}")
            else:
                parts_for_format.append(part)

        return parts_for_format

    def __str__(self):
        return '.'.join(self.parts_for_format())


class ComplexIdentWithPrefixAndArgs(ComplexIdentWithPrefix):
    def __init__(self, env_prefix, *parts, data_types: List[BaseDataType]):
        self.env_prefix = env_prefix
        self.parts = tuple(self.validate_ident(part) for part in parts)
        self.data_types = data_types

    def __str__(self):
        return f"{super().__str__()}({','.join(data_type.name for data_type in self.data_types)})"


class ComplexIdentWithPrefixAndPath(ComplexIdentWithPrefix):
    def __init__(self, env_prefix, *parts, path: str):
        self.env_prefix = env_prefix
        self.parts = tuple(self.validate_ident(part) for part in parts)
        self.path = path

    def __str__(self):
        return f"{super().__str__()}({self.path})"



from setuptools import find_packages, setup
from pathlib import Path

__version__ = None
version_path = Path(__file__).parent / 'snowddl' / 'version.py'
exec(version_path.read_text())

setup(
    name='snowddl',
    version=__version__,
    description='Object management automation tool for Snowflake',
    long_description="""
        SnowDDL is a [declarative-style](https://www.snowflake.com/blog/embracing-agile-software-delivery-and-devops-with-snowflake/) tool for object management automation in Snowflake.

        Getting started: [https://docs.snowddl.com/getting-started](https://docs.snowddl.com/getting-started)

        Main features: [https://docs.snowddl.com/features](https://docs.snowddl.com/features)

        1. SnowDDL is "stateless".
        2. SnowDDL can revert any changes.
        3. SnowDDL supports ALTER COLUMN.
        4. SnowDDL provides built-in "Role hierarchy" model.
        5. SnowDDL re-creates invalid views automatically.
        6. SnowDDL simplifies code review.
        7. SnowDDL supports creation of isolated "environments" for individual developers and CI/CD scripts.
        8. SnowDDL strikes a good balance between dependency management overhead and parallelism.
        9. SnowDDL configuration can be generated dynamically in Python code.
        10. SnowDDL can manage packages for Java and Python UDF scripts natively.

        Enjoy!
    """,
    long_description_content_type='text/markdown',
    url='https://github.com/littleK0i/snowddl',
    author='Vitaly Markov',
    author_email='wild.desu@gmail.com',

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Topic :: Database',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],

    keywords='snowflake database schema object change ddl sql create alter drop grant',

    packages=find_packages(),
    include_package_data = True,

    entry_points={
        'console_scripts': [
            'snowddl = snowddl.__main__:default_entry_point',
            'snowddlconv = snowddl.__main__:convert_entry_point'
        ],
    },

    install_requires=[
        'snowflake-connector-python',
        'pyyaml',
        'jsonschema',
    ],

    extras_require={
        'test': ['pytest', 'pytest-xdist'],
    },

    python_requires='>=3.7',
)



from argparse import ArgumentParser, HelpFormatter
from logging import getLogger, Formatter, StreamHandler
from os import environ, getcwd
from pathlib import Path
from snowflake.connector import connect
from shutil import rmtree

from snowddl import SnowDDLApp, SnowDDLSettings, ObjectType


def default_entry_point():
    formatter = lambda prog: HelpFormatter(prog, max_help_position=32)
    parser = ArgumentParser(prog='snowddl', description='Object management automation tool for Snowflake', formatter_class=formatter)

    # Config
    parser.add_argument('-c', help='Path to config directory OR name of bundled test config (default: current directory)', metavar='CONFIG_PATH', default=getcwd())

    # Auth
    parser.add_argument('-a', help='Snowflake account identifier (default: SNOWFLAKE_ACCOUNT env variable)', metavar='ACCOUNT', default=environ.get('SNOWFLAKE_ACCOUNT'))
    parser.add_argument('-u', help='Snowflake user name (default: SNOWFLAKE_USER env variable)', metavar='USER', default=environ.get('SNOWFLAKE_USER'))
    parser.add_argument('-p', help='Snowflake user password (default: SNOWFLAKE_PASSWORD env variable)', metavar='PASSWORD', default=environ.get('SNOWFLAKE_PASSWORD'))
    parser.add_argument('-k', help='Path to private key file (default: SNOWFLAKE_PRIVATE_KEY_PATH env variable)', metavar='PRIVATE_KEY', default=environ.get('SNOWFLAKE_PRIVATE_KEY_PATH'))

    # Role & warehouse
    parser.add_argument('-r', help='Snowflake active role (default: SNOWFLAKE_ROLE env variable)', metavar='ROLE', default=environ.get('SNOWFlAKE_ROLE'))
    parser.add_argument('-w', help='Snowflake active warehouse (default: SNOWFLAKE_WAREHOUSE env variable)', metavar='WAREHOUSE', default=environ.get('SNOWFLAKE_WAREHOUSE'))

    # Options
    parser.add_argument('--passphrase', help='Passphrase for private key file (default: SNOWFLAKE_PRIVATE_KEY_PASSPHRASE env variable)', default=environ.get('SNOWFLAKE_PRIVATE_KEY_PASSPHRASE'))
    parser.add_argument('--env-prefix', help='Env prefix added to global object names, used to separate environments (e.g. DEV, PROD)', default=None)
    parser.add_argument('--max-workers', help='Maximum number of workers to resolve objects in parallel', default=None, type=int)
    parser.add_argument('--placeholder-path', help='Path to config file with environment-specific placeholders', default=None)

    # Logging
    parser.add_argument('--log-level', help="Log level (possible values: DEBUG, INFO, WARNING; default: INFO)", default="INFO")
    parser.add_argument('--show-sql', help="Show executed DDL queries", default=False, action='store_true')

    # Object types
    parser.add_argument('--exclude-object-types', help="Comma-separated list of object types NOT to resolve", default=None, metavar='')
    parser.add_argument('--include-object-types', help="Comma-separated list of object types TO resolve, all other types are excluded", default=None, metavar='')

    # Apply even more unsafe changes
    parser.add_argument('--apply-unsafe', help="Additionally apply unsafe changes, which may cause loss of data (ALTER, DROP, etc.)", default=False, action='store_true')
    parser.add_argument('--apply-replace-table', help="Additionally apply REPLACE TABLE when ALTER TABLE is not possible", default=False, action='store_true')
    parser.add_argument('--apply-masking-policy', help="Additionally apply changes to MASKING POLICIES", default=False, action='store_true')
    parser.add_argument('--apply-row-access-policy', help="Additionally apply changes to ROW ACCESS POLICIES", default=False, action='store_true')
    parser.add_argument('--apply-account-params', help="Additionally apply changes to ACCOUNT PARAMETERS", default=False, action='store_true')
    parser.add_argument('--apply-network-policy', help="Additionally apply changes to NETWORK POLICIES", default=False, action='store_true')
    parser.add_argument('--apply-resource-monitor', help="Additionally apply changes to RESOURCE MONITORS", default=False, action='store_true')

    # Refresh state of specific objects
    parser.add_argument('--refresh-user-passwords', help="Additionally refresh passwords of users", default=False, action='store_true')
    parser.add_argument('--refresh-future-grants', help="Additionally refresh missing grants for existing objects derived from future grants", default=False, action='store_true')

    # Destroy without env prefix
    parser.add_argument('--destroy-without-prefix', help="Allow {destroy} action without --env-prefix", default=False, action='store_true')

    # Subparsers
    subparsers = parser.add_subparsers(dest="action")
    subparsers.required = True

    subparsers.add_parser('plan', help="Resolve objects, apply nothing, display suggested changes")
    subparsers.add_parser('apply', help="Resolve objects, apply safe changes, display suggested unsafe changes")
    subparsers.add_parser('destroy', help="Drop objects with specified --env-prefix, use it to reset dev and test environments")

    args = parser.parse_args()

    if not args.a or not args.u or (not args.p and not args.k):
        parser.print_help()
        exit(1)

    logger = get_logger(args)
    config_path = get_config_path(args)
    placeholder_path = get_placeholder_path(args)

    app = SnowDDLApp()
    app.init_config(args.env_prefix)
    app.init_engine(get_connection(args), get_engine_settings(args))

    app.output_context()

    app.load_placeholders_with_parsers(config_path, placeholder_path)

    if app.config.errors:
        app.output_config_errors()
        exit(1)

    app.load_blueprints_with_parsers(config_path)

    if app.config.errors:
        app.output_config_errors()
        exit(1)

    if args.action == 'destroy':
        if not args.env_prefix and not args.destroy_without_prefix:
            raise ValueError("Argument --env-prefix is required for [destroy] action")

        app.destroy_objects()
    else:
        app.resolve_objects()

    app.output_engine_stats()

    if args.show_sql:
        app.output_executed_ddl()

    app.output_suggested_ddl()


def convert_entry_point():
    formatter = lambda prog: HelpFormatter(prog, max_help_position=32)
    parser = ArgumentParser(prog='snowddlconv', description='Convert existing objects in Snowflake account to SnowDDL config', formatter_class=formatter)

    # Config
    parser.add_argument('-c', help='Path to output config directory (default: current directory)', metavar='CONFIG_PATH', default=getcwd())

    # Auth
    parser.add_argument('-a', help='Snowflake account identifier (default: SNOWFLAKE_ACCOUNT env variable)', metavar='ACCOUNT', default=environ.get('SNOWFLAKE_ACCOUNT'))
    parser.add_argument('-u', help='Snowflake user name (default: SNOWFLAKE_USER env variable)', metavar='USER', default=environ.get('SNOWFLAKE_USER'))
    parser.add_argument('-p', help='Snowflake user password (default: SNOWFLAKE_PASSWORD env variable)', metavar='PASSWORD', default=environ.get('SNOWFLAKE_PASSWORD'))
    parser.add_argument('-k', help='Path to private key file (default: SNOWFLAKE_PRIVATE_KEY_PATH env variable)', metavar='PRIVATE_KEY', default=environ.get('SNOWFLAKE_PRIVATE_KEY_PATH'))

    # Role & warehouse
    parser.add_argument('-r', help='Snowflake active role (default: SNOWFLAKE_ROLE env variable)', metavar='ROLE', default=environ.get('SNOWFlAKE_ROLE'))
    parser.add_argument('-w', help='Snowflake active warehouse (default: SNOWFLAKE_WAREHOUSE env variable)', metavar='WAREHOUSE', default=environ.get('SNOWFLAKE_WAREHOUSE'))

    # Options
    parser.add_argument('--passphrase', help='Passphrase for private key file (default: SNOWFLAKE_PRIVATE_KEY_PASSPHRASE env variable)', default=environ.get('SNOWFLAKE_PRIVATE_KEY_PASSPHRASE'))
    parser.add_argument('--env-prefix', help='Env prefix added to global object names, used to separate environments (e.g. DEV, PROD)', default=None)
    parser.add_argument('--max-workers', help='Maximum number of workers to resolve objects in parallel', default=None, type=int)
    parser.add_argument('--clean', help='Delete existing config files before conversion', default=False, action='store_true')

    # Logging
    parser.add_argument('--log-level', help="Log level (possible values: DEBUG, INFO, WARNING; default: INFO)", default="INFO")

    # Object types
    parser.add_argument('--exclude-object-types', help="Comma-separated list of object types NOT to convert", default=None, metavar='')
    parser.add_argument('--include-object-types', help="Comma-separated list of object types TO convert, all other types are excluded", default=None, metavar='')

    args = parser.parse_args()

    if not args.a or not args.u or (not args.p and not args.k):
        parser.print_help()
        exit(1)

    logger = get_logger(args)
    config_path = get_convert_config_path(args)

    app = SnowDDLApp()
    app.init_config(args.env_prefix)
    app.init_engine(get_connection(args), get_convert_engine_settings(args))

    app.output_context()
    app.convert_objects(config_path)


def get_config_path(args):
    config_path = Path(args.c)

    if not config_path.exists():
        config_path = Path(__file__).parent / '_config' / args.c

    if not config_path.exists():
        raise ValueError(f"Config path [{args.c}] does not exist")

    if not config_path.is_dir():
        raise ValueError(f"Config path [{args.c}] is not a directory")

    return config_path.resolve()


def get_placeholder_path(args):
    if args.placeholder_path:
        placeholder_path = Path(args.placeholder_path)

        if not placeholder_path.is_file():
            raise ValueError(f"Placeholder path [{args.placeholder_path}] does not exist or not a file")

        return placeholder_path.resolve()

    return None

def get_convert_config_path(args):
    config_path = Path(args.c)

    if config_path.is_dir() and args.clean:
        rmtree(config_path)

    if not config_path.exists():
        config_path.mkdir(mode=0o755, parents=True)

    if not config_path.is_dir():
        raise ValueError(f"Config path [{args.c}] is not a directory")

    return config_path


def get_logger(args):
    logger = getLogger('snowddl')
    logger.setLevel(args.log_level)

    formatter = Formatter('%(asctime)s - %(levelname)s - %(message)s')
    formatter.default_msec_format = '%s.%03d'

    handler = StreamHandler()
    handler.setFormatter(formatter)

    logger.addHandler(handler)

    return logger


def get_connection(args):
    options = {
        "account": args.a,
        "user": args.u,
        "role": args.r,
        "warehouse": args.w,
    }

    if args.k:
        from cryptography.hazmat.primitives import serialization

        key_path = Path(args.k)
        key_password = args.passphrase.encode() if args.passphrase else None

        pk = serialization.load_pem_private_key(data=key_path.read_bytes(), password=key_password)

        options['private_key'] =  pk.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
    else:
        options['password'] = args.p

    return connect(**options)


def get_engine_settings(args):
    settings = SnowDDLSettings()

    if args.action == 'apply' or args.action == 'destroy':
        settings.execute_safe_ddl = True

        if args.apply_unsafe:
            settings.execute_unsafe_ddl = True

        if args.apply_replace_table:
            settings.execute_replace_table = True

        if args.apply_masking_policy:
            settings.execute_masking_policy = True

        if args.apply_row_access_policy:
            settings.execute_row_access_policy = True

        if args.apply_account_params:
            settings.execute_account_params = True

        if args.apply_network_policy:
            settings.execute_network_policy = True

        if args.apply_resource_monitor:
            settings.execute_resource_monitor = True

    if args.refresh_user_passwords:
        settings.refresh_user_passwords = True

    if args.refresh_future_grants:
        settings.refresh_future_grants = True

    if args.exclude_object_types:
        try:
            settings.exclude_object_types = [ObjectType[t.strip().upper()] for t in str(args.exclude_object_types).split(',')]
        except KeyError as e:
            raise ValueError(f"Invalid object type [{str(e)}]")

    if args.include_object_types:
        try:
            settings.include_object_types = [ObjectType[t.strip().upper()] for t in str(args.include_object_types).split(',')]
        except KeyError as e:
            raise ValueError(f"Invalid object type [{str(e)}]")

    if args.max_workers:
        settings.max_workers = args.max_workers

    return settings


def get_convert_engine_settings(args):
    settings = SnowDDLSettings()

    if args.exclude_object_types:
        try:
            settings.exclude_object_types = [ObjectType[t.strip().upper()] for t in str(args.exclude_object_types).split(',')]
        except KeyError as e:
            raise ValueError(f"Invalid object type [{str(e)}]")

    if args.include_object_types:
        try:
            settings.include_object_types = [ObjectType[t.strip().upper()] for t in str(args.include_object_types).split(',')]
        except KeyError as e:
            raise ValueError(f"Invalid object type [{str(e)}]")

    if args.max_workers:
        settings.max_workers = args.max_workers

    return settings


if __name__ == '__main__':
    default_entry_point()



from snowddl.blueprint import PipeBlueprint, Ident, IdentWithPrefix, ComplexIdentWithPrefix
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


pipe_json_schema = {
    "type": "object",
    "properties": {
        "copy": {
            "type": "object",
            "properties": {
                "table": {
                    "type": "string"
                },
                "stage": {
                    "type": "string"
                },
                "path": {
                    "type": "string"
                },
                "pattern": {
                    "type": "string"
                },
                "file_format": {
                    "type": "string"
                },
                "transform": {
                    "type": "object",
                    "additionalProperties": {
                        "type": "string"
                    }
                },
                "options": {
                    "type": "object",
                    "additionalProperties": {
                        "type": ["array", "boolean", "number", "string"]
                    }
                }
            },
            "required": ["table", "stage"],
            "additionalProperties": False
        },
        "auto_ingest": {
            "type": "boolean"
        },
        "aws_sns_topic": {
            "type": "string"
        },
        "integration": {
            "type": "string"
        },
        "comment": {
            "type": "string"
        }
    },
    "required": ["copy", "auto_ingest"],
    "additionalProperties": False
}


class PipeParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("pipe", pipe_json_schema, self.process_pipe)

    def process_pipe(self, f: ParsedFile):
        copy = f.params['copy']

        bp = PipeBlueprint(
            full_name=ComplexIdentWithPrefix(self.env_prefix, f.database, f.schema, f.name),
            database=IdentWithPrefix(self.env_prefix, f.database),
            schema=Ident(f.schema),
            name=Ident(f.name),
            auto_ingest=f.params['auto_ingest'],
            copy_table_name=self.config.build_complex_ident(copy['table'], f.database, f.schema),
            copy_stage_name=self.config.build_complex_ident(copy['stage'], f.database, f.schema),
            copy_path=copy.get('path'),
            copy_pattern=copy.get('pattern'),
            copy_transform=self.normalise_params_dict(copy.get('transform')),
            copy_file_format=self.config.build_complex_ident(copy.get('file_format'), f.database, f.schema) if copy.get('file_format') else None,
            copy_options=self.normalise_params_dict(copy.get('options')),
            aws_sns_topic=f.params.get('aws_sns_topic'),
            integration=Ident(f.params['integration']) if f.params.get('integration') else None,
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)



from abc import abstractmethod

from snowddl.blueprint import SchemaBlueprint
from snowddl.resolver.abc_resolver import AbstractResolver, ResolveResult, ObjectType


class AbstractSchemaObjectResolver(AbstractResolver):
    def get_existing_objects(self):
        existing_objects = {}

        # Process schemas in parallel
        for schema_objects in self.engine.executor.map(self.get_existing_objects_in_schema, self.engine.schema_cache.schemas.values()):
            existing_objects.update(schema_objects)

        return existing_objects

    def destroy(self):
        # Do nothing, all schema objects are dropped automatically on DROP DATABASE
        pass

    @abstractmethod
    def get_existing_objects_in_schema(self, schema: dict):
        pass

    def _resolve_drop(self):
        # Drop existing objects without blueprints
        # with additional check for "sandbox" schema
        tasks = {}

        for full_name in sorted(self.existing_objects):
            if full_name not in self.blueprints and not self._is_sandbox_schema(full_name):
                tasks[full_name] = (self.drop_object, self.existing_objects[full_name])

        self._process_tasks(tasks)

    def _is_sandbox_schema(self, object_full_name):
        schema_full_name = '.'.join(object_full_name.split('.')[:2])
        schema_bp = self.config.get_blueprints_by_type(SchemaBlueprint).get(schema_full_name)

        if schema_bp and schema_bp.is_sandbox:
            return True

        return False



from snowddl.blueprint import SequenceBlueprint, Ident, IdentWithPrefix, ComplexIdentWithPrefix
from snowddl.parser.abc_parser import AbstractParser, ParsedFile


sequence_json_schema = {
    "type": "object",
    "properties": {
        "start": {
            "type": "integer"
        },
        "interval": {
            "type": "integer"
        },
        "comment": {
            "type": "string"
        }
    },
    "additionalProperties": False
}


class SequenceParser(AbstractParser):
    def load_blueprints(self):
        self.parse_schema_object_files("sequence", sequence_json_schema, self.process_sequence)

    def process_sequence(self, f: ParsedFile):
        bp = SequenceBlueprint(
            full_name=ComplexIdentWithPrefix(self.env_prefix, f.database, f.schema, f.name),
            database=IdentWithPrefix(self.env_prefix, f.database),
            schema=Ident(f.schema),
            name=Ident(f.name),
            start=f.params.get('start', 1),
            interval=f.params.get('interval', 1),
            comment=f.params.get('comment'),
        )

        self.config.add_blueprint(bp)



from itertools import islice
from re import compile

from snowddl.blueprint import TableBlueprint, TableColumn, DataType, BaseDataType
from snowddl.resolver.abc_schema_object_resolver import AbstractSchemaObjectResolver, ResolveResult, ObjectType

cluster_by_syntax_re = compile(r'^(\w+)?\((.*)\)$')


class TableResolver(AbstractSchemaObjectResolver):
    def get_object_type(self) -> ObjectType:
        return ObjectType.TABLE

    def get_existing_objects_in_schema(self, schema: dict):
        existing_objects = {}

        cur = self.engine.execute_meta("SHOW TABLES IN SCHEMA {database:i}.{schema:i}", {
            "database": schema['database'],
            "schema": schema['schema'],
        })

        for r in cur:
            # Skip external tables
            if r['is_external'] == 'Y':
                continue

            full_name = f"{r['database_name']}.{r['schema_name']}.{r['name']}"
            existing_objects[full_name] = {
                "database": r['database_name'],
                "schema": r['schema_name'],
                "name": r['name'],
                "owner": r['owner'],
                "is_transient": r['kind'] == 'TRANSIENT',
                "cluster_by": r['cluster_by'] if r['cluster_by'] else None,
                "change_tracking": bool(r['change_tracking'] == 'ON'),
                "search_optimization": bool(r.get('search_optimization') == 'ON'),
                "comment": r['comment'] if r['comment'] else None,
            }

        return existing_objects

    def get_blueprints(self):
        return self.config.get_blueprints_by_type(TableBlueprint)

    def create_object(self, bp: TableBlueprint):
        query = self._build_create_table(bp)
        self.engine.execute_safe_ddl(query)

        if bp.search_optimization:
            self.engine.execute_safe_ddl("ALTER TABLE {full_name:i} ADD SEARCH OPTIMIZATION", {
                "full_name": bp.full_name,
            })

        return ResolveResult.CREATE

    def compare_object(self, bp: TableBlueprint, row: dict):
        alters = []
        is_replace_required = False

        bp_cols = {str(c.name): c for c in bp.columns}
        snow_cols = self._get_existing_columns(bp)

        remaining_col_names = list(snow_cols.keys())

        for col_name, snow_c in snow_cols.items():
            # Drop columns which do not exist in blueprint
            if col_name not in bp_cols:
                alters.append(self.engine.format("DROP COLUMN {col_name:i}", {
                    "col_name": col_name,
                }))

                remaining_col_names.remove(col_name)
                continue

            bp_c = bp_cols[col_name]

            # Set or drop NOT NULL constraint
            if snow_c.not_null and not bp_c.not_null:
                alters.append(self.engine.format("MODIFY COLUMN {col_name:i} DROP NOT NULL", {
                    "col_name": col_name,
                }))
            elif not snow_c.not_null and bp_c.not_null:
                alters.append(self.engine.format("MODIFY COLUMN {col_name:i} SET NOT NULL", {
                    "col_name": col_name,
                }))

            # Default
            if snow_c.default != bp_c.default:
                # DROP DEFAULT is supported
                if snow_c.default is not None and bp_c.default is None:
                    alters.append(self.engine.format("MODIFY COLUMN {col_name:i} DROP DEFAULT", {
                        "col_name": col_name,
                    }))

                # Switch to another sequence is supported
                elif isinstance(snow_c.default, str) and snow_c.default.upper().endswith('.NEXTVAL') \
                and isinstance(bp_c.default, str) and bp_c.default.upper().endswith('.NEXTVAL'):
                    alters.append(self.engine.format("MODIFY COLUMN {col_name:i} SET DEFAULT {default:r}", {
                        "col_name": col_name,
                        "default": bp_c.default,
                    }))

                # All other DEFAULT changes are not supported
                else:
                    is_replace_required = True

            # Comments
            if snow_c.comment != bp_c.comment:
                # UNSET COMMENT is currently not supported for columns, we can only set it to empty string
                alters.append(self.engine.format("MODIFY COLUMN {col_name:i} COMMENT {comment}", {
                    "col_name": col_name,
                    "comment": bp_c.comment if bp_c.comment else '',
                }))

            # If type matches exactly, skip all other checks
            if snow_c.type == bp_c.type:
                continue

            # Only a few optimized MODIFY COLUMN ... TYPE actions are supported
            # https://docs.snowflake.com/en/sql-reference/sql/alter-table-column.html
            if snow_c.type.base_type == bp_c.type.base_type:
                # Increase or decrease precision of NUMBER, but not scale
                if snow_c.type.base_type == BaseDataType.NUMBER \
                and snow_c.type.val1 != bp_c.type.val2 \
                and snow_c.type.val2 == bp_c.type.val2:
                    alters.append(self.engine.format("MODIFY COLUMN {col_name:i} TYPE {col_type:r}", {
                        "col_name": col_name,
                        "col_type": bp_c.type,
                    }))

                    continue

                if snow_c.type.base_type == BaseDataType.VARCHAR \
                and snow_c.type.val1 < bp_c.type.val1:
                    alters.append(self.engine.format("MODIFY COLUMN {col_name:i} TYPE {col_type:r}", {
                        "col_name": col_name,
                        "col_type": bp_c.type,
                    }))

                    continue

            # All other transformations require full table replace
            is_replace_required = True

        # Remaining column names exactly match initial part of blueprint column names
        if remaining_col_names == list(islice(bp_cols.keys(), 0, len(remaining_col_names))):
            # Get remaining part of blueprint columns
            for col_name, bp_c in islice(bp_cols.items(), len(remaining_col_names), None):
                query = self.engine.query_builder()
                query.append("ADD COLUMN {col_name:i} {col_type:r}", {
                    "col_name": col_name,
                    "col_type": bp_c.type,
                })

                if bp_c.default is not None:
                    query.append("DEFAULT {default}", {
                        "default": bp_c.default,
                    })

                if bp_c.not_null:
                    query.append("NOT NULL")

                if bp_c.comment:
                    query.append("COMMENT {comment}", {
                        "comment": bp_c.comment,
                    })

                alters.append(query)
        else:
            # Reordering of columns is not supported
            is_replace_required = True

        # Changing TRANSIENT tables to permanent and back are not supported
        if bp.is_transient != row['is_transient']:
            is_replace_required = True

        # Clustering key
        if not self._compare_cluster_by(bp, row):
            if bp.cluster_by:
                alters.append(self.engine.format("CLUSTER BY ({cluster_by:r})", {
                    "cluster_by": bp.cluster_by,
                }))
            else:
                alters.append(self.engine.format("DROP CLUSTERING KEY"))

        # Change tracking
        if bp.change_tracking != row['change_tracking']:
            alters.append(self.engine.format("SET CHANGE_TRACKING = {change_tracking:b}", {
                "change_tracking": bp.change_tracking,
            }))

        # Search optimization
        if bp.search_optimization and not row['search_optimization']:
            alters.append("ADD SEARCH OPTIMIZATION")
        elif not bp.search_optimization and row['search_optimization']:
            alters.append("DROP SEARCH OPTIMIZATION")

        # Comment
        if bp.comment != row['comment']:
            if bp.comment:
                alters.append(self.engine.format("SET COMMENT = {comment}", {
                    "comment": bp.comment,
                }))
            else:
                alters.append(self.engine.format("UNSET COMMENT"))

        if is_replace_required:
            self.engine.execute_unsafe_ddl(self._build_create_table(bp, snow_cols), condition=self.engine.settings.execute_replace_table)

            if bp.search_optimization:
                self.engine.execute_safe_ddl("ALTER TABLE {full_name:i} ADD SEARCH OPTIMIZATION", {
                    "full_name": bp.full_name,
                })

            return ResolveResult.REPLACE
        elif alters:
            for alter in alters:
                self.engine.execute_unsafe_ddl("ALTER TABLE {full_name:i} {alter:r}", {
                    "full_name": bp.full_name,
                    "alter": alter,
                })

            return ResolveResult.ALTER

        return ResolveResult.NOCHANGE

    def drop_object(self, row: dict):
        self.engine.execute_unsafe_ddl("DROP TABLE {database:i}.{schema:i}.{table_name:i}", {
            "database": row['database'],
            "schema": row['schema'],
            "table_name": row['name'],
        })

        return ResolveResult.DROP

    def _get_existing_columns(self, bp: TableBlueprint):
        existing_columns = {}

        cur = self.engine.execute_meta("DESC TABLE {full_name:i}", {
            "full_name": bp.full_name,
        })

        for r in cur:
            existing_columns[r['name']] = TableColumn(
                name=r['name'],
                type=DataType(r['type']),
                not_null=bool(r['null?'] == 'N'),
                default=r['default'] if r['default'] else None,
                comment=r['comment'] if r['comment'] else None,
            )

        return existing_columns

    def _build_create_table(self, bp: TableBlueprint, snow_cols=None):
        query = self.engine.query_builder()
        query.append("CREATE")

        if snow_cols:
            query.append("OR REPLACE")

        if bp.is_transient:
            query.append("TRANSIENT")

        query.append("TABLE {full_name:i}", {
            "full_name": bp.full_name,
        })

        query.append_nl("(")

        for idx, c in enumerate(bp.columns):
            query.append_nl("    {comma:r}{col_name:i} {col_type:r}", {
                "comma": "  " if idx == 0 else ", ",
                "col_name": c.name,
                "col_type": c.type,
            })

            if c.default is not None:
                query.append("DEFAULT {default:r}", {
                    "default": c.default,
                })

            if c.not_null:
                query.append("NOT NULL")

            if c.comment:
                query.append("COMMENT {comment}", {
                    "comment": c.comment,
                })

        query.append_nl(")")

        if bp.cluster_by:
            query.append_nl("CLUSTER BY ({cluster_by:r})", {
                "cluster_by": bp.cluster_by,
            })

        if bp.change_tracking:
            query.append_nl("CHANGE_TRACKING = TRUE")

        if bp.retention_time is not None:
            query.append_nl("DATA_RETENTION_TIME_IN_DAYS = {retention_time:d}", {
                "retention_time": bp.retention_time
            })

        if bp.comment:
            query.append_nl("COMMENT = {comment}", {
                "comment": bp.comment,
            })

        if snow_cols:
            query.append_nl("COPY GRANTS")
            query.append_nl("AS")
            query.append_nl("SELECT")

            for idx, c in enumerate(bp.columns):
                if str(c.name) in snow_cols:
                    query.append_nl("    {comma:r}{col_name:i}::{col_type:r} AS {col_name:i}", {
                        "comma": "  " if idx == 0 else ", ",
                        "col_name": c.name,
                        "col_type": c.type,
                    })
                else:
                    query.append_nl("    {comma:r}{col_val}::{col_type:r} AS {col_name:i}", {
                        "comma": "  " if idx == 0 else ", ",
                        "col_name": c.name,
                        "col_type": c.type,
                        "col_val": c.default,
                    })

            query.append_nl("FROM {full_name:i}", {
                "full_name": bp.full_name,
            })

        return query

    def _compare_cluster_by(self, bp: TableBlueprint, row: dict):
        bp_cluster_by = ', '.join(bp.cluster_by) if bp.cluster_by else None
        snow_cluster_by = cluster_by_syntax_re.sub(r'\2', row['cluster_by']) if row['cluster_by'] else None

        return bp_cluster_by == snow_cluster_by



from logging import getLogger, NullHandler
from threading import get_ident

from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from snowflake.connector import DictCursor, SnowflakeConnection, Error

from snowddl.config import SnowDDLConfig
from snowddl.settings import SnowDDLSettings
from snowddl.formatter import SnowDDLFormatter
from snowddl.query_builder import SnowDDLQueryBuilder
from snowddl.context import SnowDDLContext
from snowddl.error import SnowDDLExecuteError
from snowddl.schema_cache import SnowDDLSchemaCache


logger = getLogger(__name__)
logger.addHandler(NullHandler())


class SnowDDLEngine:
    def __init__(self, connection: SnowflakeConnection, config: SnowDDLConfig, settings: SnowDDLSettings):
        self.connection = connection
        self.config = config
        self.settings = settings

        self.formatter = SnowDDLFormatter()
        self.logger = logger

        self.executor = ThreadPoolExecutor(max_workers=self.settings.max_workers, thread_name_prefix=self.__class__.__name__)

        self.executed_ddl = []
        self.suggested_ddl = []

        self._executed_ddl_buffer = defaultdict(list)
        self._suggested_ddl_buffer = defaultdict(list)

        self.context = SnowDDLContext(self)
        self.context.activate_role_with_prefix()

        self.schema_cache = SnowDDLSchemaCache(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.executor.shutdown()

    def format(self, sql, params=None):
        sql = str(sql)

        if params:
            return self.formatter.format(sql, **params)

        return sql

    def query_builder(self):
        return SnowDDLQueryBuilder(self)

    def describe_meta(self, sql, params=None):
        return self._describe(sql, params)

    def execute_meta(self, sql, params=None):
        return self._execute(sql, params, is_meta=True)

    def execute_context_ddl(self, sql, params=None):
        return self._execute(sql, params)

    def execute_safe_ddl(self, sql, params=None, condition=True, file_stream=None):
        if self.settings.execute_safe_ddl and condition:
            self._execute(sql, params, False, file_stream)
        else:
            self._suggest(sql, params)

    def execute_unsafe_ddl(self, sql, params=None, condition=True, file_stream=None):
        if self.settings.execute_unsafe_ddl and condition:
            self._execute(sql, params, False, file_stream)
        else:
            self._suggest(sql, params)

    def flush_thread_buffers(self):
        for thread_sql in self._executed_ddl_buffer.values():
            for sql in thread_sql:
                self.executed_ddl.append(sql)

        for thread_sql in self._suggested_ddl_buffer.values():
            for sql in thread_sql:
                self.suggested_ddl.append(sql)

        self._executed_ddl_buffer = defaultdict(list)
        self._suggested_ddl_buffer = defaultdict(list)

    def _execute(self, sql, params, is_meta=False, file_stream=None):
        sql = self.format(sql, params)

        try:
            result = self.connection.cursor(DictCursor).execute(sql, file_stream=file_stream)
        except Error as e:
            raise SnowDDLExecuteError(e, sql)

        if not is_meta:
            self._executed_ddl_buffer[get_ident()].append(sql)

        return result

    def _describe(self, sql, params):
        sql = self.format(sql, params)

        try:
            result = self.connection.cursor(DictCursor).describe(sql)
        except Error as e:
            raise SnowDDLExecuteError(e, sql)

        return result

    def _suggest(self, sql, params):
        sql = self.format(sql, params)
        self._suggested_ddl_buffer[get_ident()].append(sql)
